#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import os
from threading import Lock

from gnuhealth.config import get_config_dir

KNOWN_HOSTS_PATH = os.path.join(get_config_dir(), 'known_hosts')

_lock = Lock()


def _load():
    if not os.path.isfile(KNOWN_HOSTS_PATH):
        return {}
    fingerprints = {}
    with open(KNOWN_HOSTS_PATH) as known_hosts:
        for line in known_hosts:
            line = line.strip()
            try:
                host, sha1 = line.split(' ')
            except ValueError:
                host, sha1 = line, ''
            fingerprints[host] = sha1
    return fingerprints


def exists(host):
    return host in _load()


def get(host):
    return _load().get(host)


def set(host, fingerprint):
    fingerprints = _load()
    assert isinstance(host, str)
    if fingerprint:
        assert len(fingerprint) == 59  # len of formated sha1
    else:
        fingerprint = ''
    changed = fingerprint != fingerprints.get(host)
    fingerprints[host] = fingerprint
    if changed:
        _save(fingerprints)


def _save(fingerprints):
    with _lock:
        with open(KNOWN_HOSTS_PATH, 'w') as known_hosts:
            known_hosts.writelines(
                '%s %s' % (host, sha1) + os.linesep
                for host, sha1 in fingerprints.items())



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText:  2024 - Wei Zhao <wei.zhao@uclouvain.be>
# SPDX-License-Identifier: GPL-3.0-or-later
#
#
# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.


import gettext
from urllib.request import urlopen
from urllib.parse import unquote

from gi.repository import Gtk

from gnuhealth.common import common
from gnuhealth.common import file_selection
from gnuhealth.common.entry_position import reset_position
from gnuhealth.gui.window.view_form.view.form_gtk.binary \
    import Binary, BinaryMixin

_ = gettext.gettext


# The binary_dicom widget is based on the existing binary widget of
# tryton.

class BinaryDicomMixin(BinaryMixin):

    def __init__(self, view, attrs):
        """
        Initialize the BinaryDicomMixin.

        :param view: The view parameter.
        :param attrs: The attrs parameter.
        """
        super(BinaryDicomMixin, self).__init__(view, attrs)
        self.filename = attrs.get('filename')

    @property
    def filters(self):
        """
        Get the list of filters to apply when selecting files.
        Return a list of Gtk.FileFilter objects.
        """
        filter_all = Gtk.FileFilter()
        filter_all.set_name(_('All files'))
        filter_all.add_pattern("*")
        filter_dicom = Gtk.FileFilter()
        filter_dicom.set_name(_('DICOM files'))
        filter_dicom.add_pattern("*.dcm")
        filter_dicom.add_pattern("*.DCM")
        filter_dicom.add_pattern("*.zip")
        filter_dicom.add_pattern("*.gz")
        return [filter_dicom, filter_all]

    def update_buttons(self, value):
        """
        Updates the visibility of buttons based on the input value.
        """
        if value:
            # self.but_save_as.show()    # don't show "save as" button
            self.but_select.hide()
            self.but_clear.show()
        else:
            self.but_save_as.hide()
            self.but_select.show()
            self.but_clear.hide()

    def select(self, widget=None):
        """
        Selects files from the file system and sets their URIs.
        """
        if not self.field:
            return
        filenames = file_selection(
            _('Select'),
            preview=self.preview,
            filters=self.filters,
            multi=True)
        if filenames:
            uris = [filename.as_uri() for filename in filenames]
            self._set_uris(uris)

    def select_drag_data_received(self, selection):
        """
        Handle the data received when an item is
        dragged and dropped onto the widget.
        """
        if not self.field:
            return
        self._set_uris(selection.get_uris())

    def _set_uris(self, uris):
        # put the data of all files (and the length of the files)
        # into a bytearray.
        # the format is:
        #   1. the bytes 'M', 'U', 'L', 'T' (ASCII 77,85,76,84)
        #   2. 8 bytes containing the length of the file (Little Endian)
        #   3. the data of the file
        #   4. repeat steps 2 and 3 for next file
        all_data = bytearray([77, 85, 76, 84])
        for uri in uris:
            uri = unquote(uri)
            data = urlopen(uri).read()
            # size in little endian 8 bytes
            all_data.extend(len(data).to_bytes(8, byteorder='little'))
            all_data.extend(data)
        # set the content of the field in the wizard
        self.field.set_client(self.record, all_data)
        if self.filename_field:
            self.filename_field.set_client(self.record, len(all_data))


# This is the dedicated widget for uploading DICOM files.  It allows
# multiple selection. It can read the DICOM files, the zipped DICOM
# files (.zip, .gz)

class BinaryDicom(BinaryDicomMixin, Binary):
    "BinaryDicom"

    def __init__(self, view, attrs):
        """
        Initializes the BinaryDicom class with the given view and attrs.

        Parameters:
            view: The view parameter.
            attrs: The attrs parameter.

        Returns:
            None
        """
        super(BinaryDicom, self).__init__(view, attrs)

        self.widget = Gtk.HBox(spacing=0)
        self.wid_size = Gtk.Entry()
        self.wid_size.set_width_chars(self.default_width_chars)
        self.wid_size.set_alignment(1.0)
        self.wid_size.props.sensitive = False
        if self.filename and attrs.get('filename_visible'):
            self.wid_text = Gtk.Entry()
            self.wid_text.set_property('activates_default', True)
            self.wid_text.connect('focus-out-event',
                                  lambda x,
                                  y: self._focus_out())
            self.wid_text.connect_after('key_press_event', self.sig_key_press)
            self.wid_text.connect('icon-press', self.sig_icon_press)
            self.widget.pack_start(
                self.wid_text, expand=True, fill=True, padding=0)
        else:
            self.wid_text = None
        self.mnemonic_widget = self.wid_text
        self.widget.pack_start(
            self.wid_size, expand=not self.filename, fill=True, padding=0)

        self.widget.pack_start(
            self.toolbar(), expand=False, fill=False, padding=0)

    def sig_icon_press(self, widget, icon_pos):
        """
        This function handles the press event for the signal icon.
        It takes in the widget, icon position, and event as parameters.
        """
        widget.grab_focus()
        if icon_pos == Gtk.EntryIconPosition.PRIMARY:
            self.open_()

    def display(self):
        """
        Displays the BinaryDicom object on the screen.
        """
        super(BinaryDicom, self).display()
        if not self.field:
            if self.wid_text:
                self.wid_text.set_text('')
            self.wid_size.set_text('')
            self.but_save_as.hide()
            return False
        if hasattr(self.field, 'get_size'):
            size = self.field.get_size(self.record)
        else:
            size = len(self.field.get(self.record))
        self.wid_size.set_text(common.humanize(size or 0))
        reset_position(self.wid_size)
        if self.wid_text:
            self.wid_text.set_text(self.filename_field.get(self.record) or '')
            reset_position(self.wid_text)
            if size:
                icon, tooltip = 'gnuhealth-open', _("Open...")
            else:
                icon, tooltip = None, ''
            pos = Gtk.EntryIconPosition.PRIMARY
            if icon:
                pixbuf = common.IconFactory.get_pixbuf(
                    icon, Gtk.IconSize.MENU)
            else:
                pixbuf = None
            self.wid_text.set_icon_from_pixbuf(pos, pixbuf)
            self.wid_text.set_icon_tooltip_text(pos, tooltip)
        self.update_buttons(bool(size))
        return True



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
"Preference"
import copy
import gettext

from gi.repository import Gdk, Gtk

import gnuhealth.rpc as rpc
from gnuhealth.common import IconFactory, RPCException, RPCExecute
from gnuhealth.common.underline import set_underline
from gnuhealth.config import GNUHEALTH_ICON
from gnuhealth.gui import Main
from gnuhealth.gui.window.nomodal import NoModal
from gnuhealth.gui.window.view_form.screen import Screen

_ = gettext.gettext


class Preference(NoModal):
    "Preference window"

    def __init__(self, user, callback):
        NoModal.__init__(self)
        self.callback = callback
        self.win = Gtk.Dialog(
            title=_('Preferences'), transient_for=self.parent,
            destroy_with_parent=True)
        Main().add_window(self.win)
        self.win.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.win.set_icon(GNUHEALTH_ICON)

        self.accel_group = Gtk.AccelGroup()
        self.win.add_accel_group(self.accel_group)

        self.but_cancel = self.win.add_button(
            set_underline(_("Cancel")), Gtk.ResponseType.CANCEL)
        self.but_cancel.set_image(
            IconFactory.get_image('gnuhealth-cancel', Gtk.IconSize.BUTTON))
        self.but_cancel.set_always_show_image(True)
        self.but_ok = self.win.add_button(
            set_underline(_("OK")), Gtk.ResponseType.OK)
        self.but_ok.set_image(
            IconFactory.get_image('gnuhealth-ok', Gtk.IconSize.BUTTON))
        self.but_ok.set_always_show_image(True)
        self.but_ok.add_accelerator(
            'clicked', self.accel_group, Gdk.KEY_Return,
            Gdk.ModifierType.CONTROL_MASK, Gtk.AccelFlags.VISIBLE)

        self.win.set_default_response(Gtk.ResponseType.OK)
        self.win.connect('response', self.response)

        try:
            view = RPCExecute('model', 'res.user',
                              'get_preferences_fields_view')
        except RPCException:
            self.win.destroy()
            self.win = None
            return

        title = Gtk.Label(label=_('Edit User Preferences'))
        title.show()
        self.win.vbox.pack_start(title, expand=False, fill=True, padding=0)
        self.screen = Screen('res.user', mode=[])
        # Reset readonly set automaticly by MODELACCESS
        self.screen.readonly = False
        self.screen.group.readonly = False
        self.screen.group.skip_model_access = True
        self.screen.add_view(view)
        self.screen.switch_view()
        self.screen.new(default=False)

        try:
            preferences = RPCExecute('model', 'res.user', 'get_preferences',
                                     False)
        except RPCException:
            self.win.destroy()
            self.win = None
            return
        self.screen.current_record.cancel()
        self.screen.current_record.set(preferences)
        self.screen.current_record.id = rpc._USER
        self.screen.current_record.validate(softvalidation=True)
        self.screen.display(set_cursor=True)

        self.screen.widget.show()
        self.win.vbox.pack_start(
            self.screen.widget, expand=True, fill=True, padding=0)
        self.win.set_title(_('Preference'))

        self.win.set_default_size(*self.default_size())

        self.register()
        self.win.show()

    def response(self, win, response_id):
        if response_id == Gtk.ResponseType.OK:
            if self.screen.current_record.validate():
                vals = copy.copy(self.screen.get())
                try:
                    RPCExecute('model', 'res.user', 'set_preferences', vals)
                except RPCException:
                    return
                rpc.context_reset()
        self.parent.present()
        self.destroy()
        self.callback()

    def destroy(self):
        self.screen.destroy()
        self.win.destroy()
        NoModal.destroy(self)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import logging
from itertools import chain, cycle, islice

from gi.repository import Gdk, GLib, Gtk

from gnuhealth.common import MODELACCESS
from gnuhealth.common.datetime_ import Date, Time

_ = gettext.gettext
logger = logging.getLogger(__name__)


def focusable_cells(column, editable=True):
    for cell in column.get_cells():
        if (not editable
                or (isinstance(cell, (
                    Gtk.CellRendererText,
                    Gtk.CellRendererCombo))
                    and cell.get_property('editable'))
                or (isinstance(cell, Gtk.CellRendererToggle)
                    and cell.get_property('activatable'))):
            yield cell


class TreeView(Gtk.TreeView):
    display_counter = 0

    def __init__(self, view):
        super(TreeView, self).__init__()
        self.view = view

    def next_column(
            self, path, column=None, cell=None, editable=True, _sign=1):
        if cell:
            cells = list(focusable_cells(column, editable))
            if len(cells) > 1:
                idx = cells.index(cell) + _sign
                if 0 <= idx < len(cells):
                    return (column, cells[idx])
        columns = self.get_columns()
        if column is None:
            column = columns[-1 if _sign > 0 else 0]
        model = self.get_model()
        record = model.get_value(model.get_iter(path), 0)
        if _sign < 0:
            columns.reverse()
        current_idx = columns.index(column) + 1
        for column in islice(cycle(columns), current_idx,
                             len(columns) + current_idx):
            if not column.name:
                continue
            widget = self.view.get_column_widget(column)
            try:
                record.load(column.name, process_exception=False)
            except Exception:
                logger.error(
                    f"Error loading '{column.name}' for {record}",
                    exc_info=True)
                return (None, None)
            field = record.group.fields[column.name]
            field.state_set(record, states=('readonly', 'invisible'))
            invisible = field.get_state_attrs(record).get('invisible', False)
            if not column.get_visible():
                invisible = True
            if editable:
                readonly = widget.attrs.get(
                    'readonly',
                    field.get_state_attrs(record).get('readonly', False))
            else:
                readonly = False
            if not (invisible or readonly):
                cells = list(focusable_cells(column, editable))
                if cells:
                    cell = cells[0 if _sign > 0 else -1]
                else:
                    continue
                return (column, cell)
        return (None, None)

    def prev_column(self, path, column=None, cell=None, editable=True):
        return self.next_column(
            path, column=column, cell=cell, editable=editable, _sign=-1)


class EditableTreeView(TreeView):
    leaving_record_events = (
        Gdk.KEY_Up, Gdk.KEY_Down, Gdk.KEY_Return)
    leaving_events = leaving_record_events + (
        Gdk.KEY_Tab, Gdk.KEY_ISO_Left_Tab, Gdk.KEY_KP_Enter)

    def on_quit_cell(
            self, current_record, column, renderer, value, callback=None):
        current_record.load(column.name, process_exception=False)
        field = current_record.group.fields[column.name]
        widget = self.view.get_column_widget(column)

        # The value has not changed and is valid ... do nothing.
        if value == widget.get_textual_value(current_record) \
                and field.validate(current_record):
            if callback:
                callback()
            return
        if widget.renderer != renderer:
            for widget in chain(widget.prefixes, widget.suffixes):
                if widget.renderer == renderer:
                    break
            else:
                raise ValueError("Unknown renderer")
        widget.value_from_text(current_record, value, callback=callback)

    def on_open_remote(self, current_record, column, create, value,
                       entry=None, callback=None):
        widget = self.view.get_column_widget(column)
        if value != widget.get_textual_value(current_record) or not value:
            changed = True
        else:
            changed = False
        try:
            widget.open_remote(current_record, create, changed, value,
                               callback=callback)
        except NotImplementedError:
            pass

    def on_create_line(self):
        access = MODELACCESS[self.view.screen.model_name]
        model = self.get_model()
        limit = (self.view.screen.size_limit is not None
                 and (len(model) >= self.view.screen.size_limit >= 0))
        if not self.view.creatable or not access['create'] or limit:
            return
        if self.view.screen.new_position == 0:
            method = model.prepend
        else:
            method = model.append
        new_record = model.group.new()
        res = method(new_record)
        sequence = self.view.attributes.get('sequence')
        if sequence:
            model.group.set_sequence(
                field=sequence, position=self.view.screen.new_position)
        return res

    def set_cursor(
            self, path, focus_column=None, cell=None, start_editing=False):
        self.grab_focus()
        if focus_column:
            widget = self.view.get_column_widget(focus_column)
            if isinstance(widget.renderer, Gtk.CellRendererToggle):
                start_editing = False
        self.scroll_to_cell(path, focus_column, use_align=False)
        if cell:
            self.set_cursor_on_cell(path, focus_column, cell, start_editing)
        else:
            super(EditableTreeView, self).set_cursor(
                path, focus_column, start_editing)

    def set_value(self):
        path, column = self.get_cursor()
        if not path or not column or not column.name:
            return True
        for renderer in column.get_cells():
            if renderer.props.editing:
                widget = self.view.get_column_widget(column)
                editable = widget.get_editable(renderer)
                self.on_editing_done(editable, renderer)
        return True

    def on_keypressed(self, entry, event, renderer):
        path = self.get_cursor()[0]
        column = self.get_column_from_renderer(renderer)
        model = self.get_model()
        record = model.get_value(model.get_iter(path), 0)
        self.display_counter += 1  # Force a display

        leaving = False
        if event.keyval == Gdk.KEY_Right:
            if isinstance(entry, Gtk.Entry):
                if entry.get_position() >= \
                        len(entry.get_text()) \
                        and not entry.get_selection_bounds():
                    leaving = True
            else:
                leaving = True
        elif event.keyval == Gdk.KEY_Left:
            if isinstance(entry, Gtk.Entry):
                if entry.get_position() <= 0 \
                        and not entry.get_selection_bounds():
                    leaving = True
            else:
                leaving = True

        if event.keyval in self.leaving_events or leaving:
            if isinstance(entry, (Date, Time)):
                entry.activate()
                txt = entry.props.value
            elif isinstance(entry, Gtk.Entry):
                txt = entry.get_text()
            elif isinstance(entry, Gtk.ComboBox):
                active = entry.get_active()
                if active < 0:
                    txt = None
                else:
                    model = entry.get_model()
                    index = entry.get_property('entry-text-column')
                    txt = model[active][index]
            else:
                return True
            keyval = event.keyval
            entry.handler_block(entry.editing_done_id)

            def callback():
                entry.handler_unblock(entry.editing_done_id)
                if (keyval in [Gdk.KEY_Tab, Gdk.KEY_KP_Enter]
                        or (keyval == Gdk.KEY_Right and leaving)):
                    GLib.idle_add(
                        self.set_cursor, path,
                        *self.next_column(path, column, renderer), True)
                elif (keyval == Gdk.KEY_ISO_Left_Tab
                        or (keyval == Gdk.KEY_Left and leaving)):
                    GLib.idle_add(
                        self.set_cursor, path,
                        *self.prev_column(path, column, renderer), True)
                elif keyval in self.leaving_record_events:
                    fields = list(self.view.widgets.keys())
                    if not record.validate(fields):
                        invalid_fields = record.invalid_fields
                        col = None
                        for col in self.get_columns():
                            if col.name in invalid_fields:
                                break
                        GLib.idle_add(self.set_cursor, path, col, None, True)
                        return
                    if ((
                        self.view.screen.pre_validate
                        and not record.pre_validate())
                            or (not self.view.screen.parent
                                and not record.save())):
                        GLib.idle_add(
                            self.set_cursor, path, column, None, True)
                        return
                    entry.handler_block(entry.editing_done_id)
                    if keyval == Gdk.KEY_Up:
                        self._key_up(path, model, column)
                    elif keyval == Gdk.KEY_Down:
                        self._key_down(path, model, column)
                    elif keyval == Gdk.KEY_Return:
                        if self.view.screen.new_position == 0:
                            new_path = self._key_up(path, model)
                        else:
                            new_path = self._key_down(path, model)
                        GLib.idle_add(self.set_cursor, new_path,
                                      *self.next_column(new_path), True)
                    entry.handler_unblock(entry.editing_done_id)
                else:
                    GLib.idle_add(self.set_cursor, path, column, None, True)
            self.on_quit_cell(record, column, renderer, txt, callback=callback)
            return True
        elif event.keyval in [Gdk.KEY_F3, Gdk.KEY_F2]:
            if isinstance(entry, Gtk.Entry):
                value = entry.get_text()
            elif isinstance(entry, Gtk.ComboBox):
                active = entry.get_active()
                if active < 0:
                    value = None
                else:
                    model = entry.get_model()
                    index = entry.get_property('entry-text-column')
                    value = model[active][index]
            else:
                return True
            entry.handler_block(entry.editing_done_id)

            def callback():
                widget = self.view.get_column_widget(column)
                value = widget.get_textual_value(record)
                if isinstance(entry, Gtk.Entry):
                    entry.set_text(value)
                else:
                    entry.set_active_text(value)
                entry.handler_unblock(entry.editing_done_id)
            self.on_open_remote(
                record, column,
                create=(event.keyval == Gdk.KEY_F3),
                value=value, callback=callback)
        else:
            field = record.group.fields[column.name]
            if isinstance(entry, Gtk.Entry):
                entry.set_max_length(int(field.attrs.get('size', 0)))
            record.modified_fields.setdefault(column.name)
            return False

        return True

    def _key_down(self, path, model, column=None):
        if path[0] == len(model) - 1 and self.view.screen.new_position == -1:
            self.on_create_line()
        new_path = Gtk.TreePath((path[0] + 1) % len(model))
        if not column:
            column, cell = self.next_column(new_path)
        self.set_cursor(new_path, column, start_editing=True)
        self.scroll_to_cell(new_path)
        return new_path

    def _key_up(self, path, model, column=None):
        if path[0] == 0 and self.view.screen.new_position == 0:
            self.on_create_line()
            new_path = Gtk.TreePath(0)
        else:
            new_path = Gtk.TreePath((path[0] - 1) % len(model))
        if not column:
            column, cell = self.next_column(new_path)
        self.set_cursor(new_path, column, start_editing=True)
        self.scroll_to_cell(new_path)
        return new_path

    def on_editing_done(self, entry, renderer):
        path = self.get_cursor()[0]
        if not path:
            return True
        column = self.get_column_from_renderer(renderer)
        model = self.get_model()
        record = model.get_value(model.get_iter(path), 0)
        if isinstance(entry, (Date, Time)):
            entry.activate()
            text = entry.props.value
        elif isinstance(entry, Gtk.ComboBox):
            model = entry.get_model()
            iter_ = entry.get_active_iter()
            if iter_:
                text = model.get_value(iter_, entry.props.entry_text_column)
            else:
                text = ''
        else:
            text = entry.get_text()
        self.on_quit_cell(record, column, renderer, text)

    def get_column_from_renderer(self, renderer):
        for column in self.get_columns():
            for cell in column.get_cells():
                if cell == renderer:
                    return column



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import configparser
import gettext
import locale
import logging
import optparse
import os
import shutil
import sys
import tempfile

from gi.repository import GdkPixbuf

from gnuhealth import __version__

logger = logging.getLogger(__name__)
_ = gettext.gettext

# Create the GNU Health plugins directory at $HOME
GH_PLUGINS_DIR = os.path.join(os.environ['HOME'], 'gnuhealth_plugins')
if not os.path.isdir(GH_PLUGINS_DIR):
    os.makedirs(GH_PLUGINS_DIR, 0o700)


def _reverse_series_iterator(starting_version):
    major, minor = map(int, starting_version.split('.'))
    while major >= 0 and minor >= 0:
        yield f"{major}.{minor}"
        if minor == 0:
            major -= 1
            minor = 8
        else:
            minor -= 2 if not minor % 2 else 1


def copy_previous_configuration(config_element):
    current_version = __version__.rsplit('.', 1)[0]
    config_dir = get_config_root()
    for version in _reverse_series_iterator(current_version):
        config_path = os.path.join(config_dir, version, config_element)
        if version == current_version and os.path.exists(config_path):
            break
        elif os.path.exists(config_path):
            if os.path.isfile(config_path):
                shutil.copy(config_path, get_config_dir())
            elif os.path.isdir(config_path):
                shutil.copytree(
                    config_path,
                    os.path.join(get_config_dir(), config_element))
            break


def get_config_root():
    config_path = os.path.expanduser(os.getenv(
        'XDG_CONFIG_HOME', os.path.join('~', '.config')))
    return os.path.join(config_path, 'gnuhealth')


def get_config_dir():
    # Allow releases in the form of major.minor.patch (X.Y.N)
    # and Release Candidate versions X.YrcN
    # Returns: X.Y
    return os.path.join(
        os.environ['HOME'], '.config', 'gnuhealth',
        '.'.join(__version__.split('.', 2)[:2]).split('rc')[0])


os.makedirs(get_config_dir(), mode=0o700, exist_ok=True)


class ConfigManager(object):
    "Config manager"

    def __init__(self):
        short_version = ''.join(__version__.split('.')[:2])
        demo_server = 'federation.gnuhealth.org:443'
        demo_database = f'health{short_version}'
        self.defaults = {
            'login.profile': demo_server,
            'login.login': 'admin',
            'login.service': '',
            'login.service.port': 8001,
            'login.host': demo_server,
            'login.db': demo_database,
            'login.expanded': True,
            'client.title': 'GNU Health HIS GTK Client',
            'client.toolbar': 'default',
            'client.save_tree_width': True,
            'client.save_tree_state': True,
            'client.spellcheck': False,
            'client.code_scanner_sound': True,
            'client.lang': locale.getdefaultlocale()[0],
            'client.language_direction': 'ltr',
            'client.email': '',
            'client.limit': 1000,
            'client.bus_timeout': 10 * 60,
            'icon.colors': '#11b0b8',
            'tree.colors': '#777,#dff0d8,#fcf8e3,#f2dede',
            'calendar.colors': '#fff,#3465a4',
            'graph.color': '#3465a4',
            'image.max_size': 10 ** 6,
            'image.cache_size': 1024,
            'doc.url': 'https://docs.gnuhealth.org/his',
            'bug.url': 'https://codeberg.org/gnuhealth/his/',
            'download.url': 'https://www.gnuhealth.org/',
            'menu.pane': 320,
        }
        self.config = {}
        self.options = {}
        self.arguments = []

    def parse(self):
        parser = optparse.OptionParser(
            version=("GNUHealth %s" % __version__),
            usage="Usage: %prog [options] [url]")
        parser.add_option(
            "-c", "--config", dest="config",
            help=_("specify alternate config file"))
        parser.add_option(
            "-d", "--dev", action="store_true",
            default=False, dest="dev", help=_("development mode"))
        parser.add_option(
            "-v", "--verbose", action="store_true",
            default=False, dest="verbose",
            help=_("logging everything at INFO level"))
        parser.add_option(
            "-l", "--log-level", dest="log_level",
            help=_("specify the log level: "
                   "DEBUG, INFO, WARNING, ERROR, CRITICAL"))
        parser.add_option(
            "-o", "--log-ouput", dest="log_output", default=None,
            help=_("specify the file used to output logging information"))
        parser.add_option(
            "-u", "--user", dest="login",
            help=_("specify the login user"))
        parser.add_option(
            "-s", "--server", dest="host",
            help=_("specify the server hostname:port"))
        parser.add_option(
            '--no-thread', default=True, action='store_false', dest='thread',
            help=_("disable thread usage"))
        opt, self.arguments = parser.parse_args()
        self.rcfile = opt.config or os.path.join(
            get_config_dir(), 'gnuhealth.conf')
        self.load()

        logging_config = {}
        if opt.log_output:
            logging_config['filename'] = opt.log_output

        loglevels = {
            'DEBUG': logging.DEBUG,
            'INFO': logging.INFO,
            'WARNING': logging.WARNING,
            'ERROR': logging.ERROR,
            'CRITICAL': logging.CRITICAL,
            }
        if not opt.log_level:
            if opt.verbose:
                opt.log_level = 'INFO'
            else:
                opt.log_level = 'ERROR'
        logging_config['level'] = loglevels[opt.log_level.upper()]
        logging.basicConfig(**logging_config)

        self.options['dev'] = opt.dev
        for arg in ['login', 'host']:
            if getattr(opt, arg):
                self.options['login.' + arg] = getattr(opt, arg)
        self.options['thread'] = opt.thread

    def save(self):
        try:
            parser = configparser.ConfigParser()
            for entry in list(self.config.keys()):
                if not len(entry.split('.')) == 2:
                    continue
                section, name = entry.split('.')
                if not parser.has_section(section):
                    parser.add_section(section)
                parser.set(section, name, str(self.config[entry]))
            with open(self.rcfile, 'w') as fp:
                parser.write(fp)
        except IOError:
            logger.warn("Unable to write config file %s", self.rcfile)
            return False
        return True

    def load(self):
        parser = configparser.ConfigParser()
        try:
            parser.read([self.rcfile])
        except configparser.Error:
            config_dir = os.path.dirname(self.rcfile)
            with tempfile.NamedTemporaryFile(
                    delete=False, prefix='gnuhealth_', suffix='.conf',
                    dir=config_dir) as temp_file:
                temp_name = temp_file.name
            shutil.copy(self.rcfile, temp_name)
            logger.error(
                f"Failed to parse {self.rcfile}. "
                f"A backup can be found at {temp_name}", exc_info=True)
            return
        for section in parser.sections():
            for (name, value) in parser.items(section):
                if value.lower() == 'true':
                    value = True
                elif value.lower() == 'false':
                    value = False
                if section == 'client' and name == 'limit':
                    # First convert to float to be backward compatible with old
                    # configuration
                    value = int(float(value))
                self.config[section + '.' + name] = value
        return True

    def __setitem__(self, key, value, config=True):
        self.options[key] = value
        if config:
            self.config[key] = value

    def __getitem__(self, key):
        return self.options.get(
            key, self.config.get(key, self.defaults.get(key)))


CONFIG = ConfigManager()
CURRENT_DIR = os.path.dirname(__file__)
if not isinstance(CURRENT_DIR, str):
    CURRENT_DIR = str(CURRENT_DIR, sys.getfilesystemencoding())

PIXMAPS_DIR = os.path.join(CURRENT_DIR, 'data', 'pixmaps', 'gnuhealth')
if not os.path.isdir(PIXMAPS_DIR):
    try:
        import importlib.resources
        ref = importlib.resources.files('gnuhealth') / 'data/pixmaps/gnuhealth'
        with importlib.resources.as_file(ref) as path:
            PIXMAPS_DIR = path
    except ImportError:
        import pkg_resources
        PIXMAPS_DIR = pkg_resources.resource_filename(
            'gnuhealth', 'data/pixmaps/gnuhealth')
SOUNDS_DIR = os.path.join(CURRENT_DIR, 'data', 'sounds')
if not os.path.isdir(SOUNDS_DIR):
    try:
        import importlib.resources
        ref = importlib.resources.files('gnuhealth') / 'data/sounds'
        with importlib.resources.as_file(ref) as path:
            SOUNDS_DIR = path
    except ImportError:
        import pkg_resources
        SOUNDS_DIR = pkg_resources.resource_filename(
            'gnuhealth', 'data/sounds')

GNUHEALTH_ICON = GdkPixbuf.Pixbuf.new_from_file(
    os.path.join(PIXMAPS_DIR, 'gnuhealth-icon.png'))

GH_ABOUT = GdkPixbuf.Pixbuf.new_from_file(
    os.path.join(PIXMAPS_DIR, 'gnuhealth-hmis-about.png'))

BANNER = 'banner.png'



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime
import gettext

from gi.repository import GLib, Gtk

from gnuhealth import common
from gnuhealth.common.datetime_ import Date as DateEntry
from gnuhealth.common.datetime_ import DateTime as DateTimeEntry
from gnuhealth.common.datetime_ import Time as TimeEntry
from gnuhealth.common.datetime_ import add_operators

from .widget import Widget

_ = gettext.gettext


class Date(Widget):
    _changed_signal = 'date-changed'

    def __init__(self, view, attrs, _entry=DateEntry):
        super(Date, self).__init__(view, attrs)

        self.widget = Gtk.HBox()
        self.entry = self.mnemonic_widget = add_operators(_entry())
        self.real_entry.set_property('activates_default', True)
        self.real_entry.connect('key_press_event', self.sig_key_press)
        self.real_entry.connect('activate', self.sig_activate)
        self.real_entry.connect('changed', lambda _: self.send_modified())
        self.real_entry.connect(
            'focus-out-event', lambda *a: self._focus_out())
        self.entry.connect(self._changed_signal, self.changed)
        self.widget.pack_start(self.entry, expand=False, fill=False, padding=0)

    @property
    def real_entry(self):
        return self.entry

    def _set_editable(self, value):
        self.entry.set_editable(value)
        self.entry.set_icon_sensitive(Gtk.EntryIconPosition.PRIMARY, value)

    def _readonly_set(self, value):
        self._set_editable(not value)

    @classmethod
    def cast(cls, value):
        if isinstance(value, datetime.datetime):
            value = value.date()
        return value

    @property
    def modified(self):
        if self.record and self.field:
            field_value = self.cast(self.field.get_client(self.record))
            return field_value != self.get_value()
        return False

    def changed(self, widget):
        def focus_out():
            if widget.props.window:
                self._focus_out()
        # Must be deferred because it triggers a display of the form
        GLib.idle_add(focus_out)

    def sig_key_press(self, widget, event):
        self.send_modified()

    def set_value(self):
        self.field.set_client(self.record, self.get_value())

    def get_value(self):
        self.entry.parse()
        return self.entry.props.value

    def set_format(self):
        if self.field and self.record:
            format_ = self.field.date_format(self.record)
        else:
            format_ = common.date_format(
                self.view.screen.context.get('date_format'))
        self.entry.props.format = format_

    def display(self):
        super(Date, self).display()
        if self.field and self.record:
            value = self.field.get_client(self.record)
        else:
            value = ''
        self.entry.props.value = value
        self.set_format()


def _move_active(combobox, scroll_type):
    if not combobox.get_child().get_editable():
        combobox.stop_emission_by_name('move-active')


class Time(Date):
    _changed_signal = 'time-changed'

    def __init__(self, view, attrs):
        super(Time, self).__init__(view, attrs, _entry=TimeEntry)
        self.entry.connect('move-active', _move_active)
        self.entry.connect(
            'scroll-event',
            lambda c, e: c.stop_emission_by_name('scroll-event'))

    def _set_editable(self, value):
        self.entry.set_sensitive(value)

    @classmethod
    def cast(cls, value):
        if isinstance(value, datetime.datetime):
            value = value.time()
        return value

    @property
    def real_entry(self):
        return self.entry.get_child()

    def display(self):
        super(Time, self).display()

    def set_format(self):
        if self.field and self.record:
            format_ = self.field.time_format(self.record)
        else:
            format_ = '%X'
        self.entry.props.format = format_


class DateTime(Date):
    _changed_signal = 'datetime-changed'

    def __init__(self, view, attrs):
        Widget.__init__(self, view, attrs)

        self.widget = Gtk.HBox()
        self.entry = self.mnemonic_widget = DateTimeEntry()
        for child in self.entry.get_children():
            add_operators(child)
            if isinstance(child, Gtk.ComboBox):
                child.connect('move-active', _move_active)
                child.connect(
                    'scroll-event',
                    lambda c, e: c.stop_emission_by_name('scroll-event'))
                child = child.get_child()
            child.set_property('activates_default', True)
            child.connect('key_press_event', self.sig_key_press)
            child.connect('activate', self.sig_activate)
            child.connect('changed', lambda _: self.send_modified())
            child.connect('focus-out-event', lambda *a: self._focus_out())
        self.entry.connect(self._changed_signal, self.changed)
        self.widget.pack_start(self.entry, expand=False, fill=False, padding=0)

    @classmethod
    def cast(cls, value):
        return value

    def _set_editable(self, value):
        for child in self.entry.get_children():
            if isinstance(child, Gtk.Entry):
                child.set_editable(value)
                child.set_icon_sensitive(Gtk.EntryIconPosition.PRIMARY, value)
            elif isinstance(child, Gtk.ComboBox):
                child.set_sensitive(value)

    def set_format(self):
        if self.field and self.record:
            date_format = self.field.date_format(self.record)
            time_format = self.field.time_format(self.record)
        else:
            date_format = common.date_format(
                self.view.screen.context.get('date_format'))
            time_format = '%X'
        self.entry.props.date_format = date_format
        self.entry.props.time_format = time_format



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import logging

from gi.repository import Gdk, Gtk

try:
    from gi.repository import GtkSpell
except ImportError:
    GtkSpell = None

from gnuhealth.config import CONFIG

from .widget import TranslateMixin, Widget

logger = logging.getLogger(__name__)


class TextBox(Widget, TranslateMixin):
    expand = True

    def __init__(self, view, attrs):
        super(TextBox, self).__init__(view, attrs)

        self.widget = Gtk.VBox()
        self.scrolledwindow = Gtk.ScrolledWindow()
        self.scrolledwindow.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        self.scrolledwindow.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        self.scrolledwindow.set_size_request(100, 100)

        self.textview = self.mnemonic_widget = self._get_textview()
        self.textview.connect('focus-out-event',
                              lambda x, y: self._focus_out())
        self.textview.connect('key-press-event', self.send_modified)
        # The click is grabbed by ListBox widget in this case user can never
        # set the input with a click
        self.textview.add_events(Gdk.EventMask.BUTTON_PRESS_MASK)
        self.textview.connect_after('button-press-event', self._button_press)
        self.scrolledwindow.add(self.textview)
        self.scrolledwindow.show_all()

        self.button = None
        if attrs.get('translate'):
            self.button = self.translate_button()
            self.widget.pack_end(
                self.button, expand=False, fill=False, padding=0)

        self.widget.pack_end(
            self.scrolledwindow, expand=True, fill=True, padding=0)

    def _get_textview(self):
        if self.attrs.get('size'):
            textbuffer = TextBufferLimitSize(int(self.attrs['size']))
            textview = Gtk.TextView()
            textview.set_buffer(textbuffer)
        else:
            textview = Gtk.TextView()
        textview.set_wrap_mode(Gtk.WrapMode.WORD)
        # TODO better tab solution
        textview.set_accepts_tab(False)
        return textview

    def _button_press(self, textview, event):
        textview.grab_focus()
        return True

    def translate_widget(self):
        box = Gtk.VBox()
        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_policy(
            Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolledwindow.set_shadow_type(Gtk.ShadowType.ETCHED_IN)
        scrolledwindow.set_size_request(-1, 80)

        textview = self._get_textview()
        scrolledwindow.add(textview)
        box.pack_end(scrolledwindow, expand=True, fill=True, padding=0)
        return box

    def translate_widget_set(self, widget, value):
        textview = widget.get_children()[-1].get_child()
        self.set_buffer(value, textview)

    def translate_widget_get(self, widget):
        textview = widget.get_children()[-1].get_child()
        return self.get_buffer(textview)

    def translate_widget_set_readonly(self, widget, value):
        textview = widget.get_children()[-1].get_child()
        textview.set_editable(not value)
        textview.props.sensitive = not value

    def _readonly_set(self, value):
        super(TextBox, self)._readonly_set(value)
        self.textview.set_editable(not value)

    @property
    def modified(self):
        if self.record and self.field:
            return self.field.get_client(self.record) != self.get_value()
        return False

    def get_value(self):
        return self.get_buffer(self.textview)

    def set_value(self):
        self.field.set_client(self.record, self.get_value())

    def set_buffer(self, value, textview):
        buf = textview.get_buffer()
        buf.delete(buf.get_start_iter(), buf.get_end_iter())
        iter_start = buf.get_start_iter()
        buf.insert(iter_start, value)

    def get_buffer(self, textview):
        buf = textview.get_buffer()
        iter_start = buf.get_start_iter()
        iter_end = buf.get_end_iter()
        return buf.get_text(iter_start, iter_end, False)

    def display(self):
        super(TextBox, self).display()
        value = self.field and self.field.get(self.record)
        if not value:
            value = ''
        self.set_buffer(value, self.textview)
        if (GtkSpell
                and self.textview.get_editable()
                and self.attrs.get('spell')
                and CONFIG['client.spellcheck']):
            checker = GtkSpell.Checker.get_from_text_view(self.textview)

            if self.record:
                language = self.record.expr_eval(self.attrs['spell'])
                if not checker:
                    checker = GtkSpell.Checker()
                    checker.attach(self.textview)
                if checker.get_language() != language:
                    try:
                        checker.set_language(language)
                    except Exception:
                        logger.debug(
                            'Could not set spell checker to "%s"', language)
                        checker.detach()
            elif checker:
                checker.detach()


class TextBufferLimitSize(Gtk.TextBuffer):
    __gsignals__ = {
        'insert-text': 'override',
    }

    def __init__(self, max_length):
        super(TextBufferLimitSize, self).__init__()
        self.max_length = max_length

    def do_insert_text(self, iter, text, length):
        free_chars = self.max_length - self.get_char_count()
        text = text[0:free_chars]
        length = len(text.encode('utf-8'))
        return Gtk.TextBuffer.do_insert_text(self, iter, text, length)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime as dt
from decimal import Decimal
from unittest import TestCase

from gnuhealth.common import untimezoned_date
from gnuhealth.common.domain_parser import (
    DomainParser, complete_value, convert_value, format_value, group_operator,
    likify, operatorize, parenthesize, quote, rlist, split_target_value, udlex)


class DomainParserTestCase(TestCase):
    "Test common domain_parser"

    def test_group_operator(self):
        "Test group_operator"
        self.assertEqual(
            list(group_operator(iter(['a', '>', '=']))), ['a', '>='])
        self.assertEqual(
            list(group_operator(iter(['>', '=', 'b']))), ['>=', 'b'])
        self.assertEqual(
            list(group_operator(iter(['a', '=', 'b']))), ['a', '=', 'b'])
        self.assertEqual(
            list(group_operator(iter(['a', '>', '=', 'b']))), ['a', '>=', 'b'])
        self.assertEqual(
            list(group_operator(iter(['a', '>', '=', '=']))), ['a', '>=', '='])

    def test_likify(self):
        "Test likify"
        for value, result in [
                ('', '%'),
                ('foo', '%foo%'),
                ('foo%', 'foo%'),
                ('foo_bar', 'foo_bar'),
                ('foo\\%', '%foo\\%%'),
                ('foo\\_bar', '%foo\\_bar%'),
        ]:
            with self.subTest(value=value):
                self.assertEqual(likify(value), result)

    def test_quote(self):
        "Test quote"
        self.assertEqual(quote('test'), 'test')
        self.assertEqual(quote('foo bar'), '"foo bar"')
        self.assertEqual(quote('"foo"'), '\\\"foo\\\"')
        self.assertEqual(quote('foo\\bar'), 'foo\\\\bar')

    def test_split_target_value(self):
        "Test split_target_value"
        field = {
            'type': 'reference',
            'selection': [
                ('spam', 'Spam'),
                ('ham', 'Ham'),
                ('e', 'Eggs'),
            ]
        }
        for value, result in (
                ('Spam', (None, 'Spam')),
                ('foo', (None, 'foo')),
                ('Spam,', ('spam', '')),
                ('Ham,bar', ('ham', 'bar')),
                ('Eggs,foo', ('e', 'foo')),
        ):
            self.assertEqual(
                split_target_value(field, value), result,
                msg="split_target_value(%r, %r)" % (field, value))

    def test_convert_boolean(self):
        "Test convert boolean"
        field = {
            'type': 'boolean',
        }
        for value, result in (
                ('Y', True),
                ('yes', True),
                ('t', True),
                ('1', True),
                ('N', False),
                ('False', False),
                ('no', False),
                ('0', False),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_float(self):
        "Test convert float"
        field = {
            'type': 'float',
        }
        for value, result in (
                ('1', 1.0),
                ('1.5', 1.5),
                ('', None),
                ('test', None),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_float_factor(self):
        "Test convert float with factor"
        field = {
            'type': 'float',
            'factor': '100',
        }
        self.assertEqual(convert_value(field, '42'), 0.42)

    def test_convert_integer(self):
        "Test convert integer"
        field = {
            'type': 'integer',
        }
        for value, result in (
                ('1', 1),
                ('1.5', 1),
                ('', None),
                ('test', None),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_integer_factor(self):
        "Test convert integer with factor"
        field = {
            'type': 'integer',
            'factor': '2',
        }
        self.assertEqual(convert_value(field, '6'), 3)

    def test_convert_numeric(self):
        "Test convert numeric"
        field = {
            'type': 'numeric',
        }
        for value, result in (
                ('1', Decimal(1)),
                ('1.5', Decimal('1.5')),
                ('', None),
                ('test', None),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_numeric_factor(self):
        "Test convert numeric with factor"
        field = {
            'type': 'numeric',
            'factor': '5',
        }
        self.assertEqual(convert_value(field, '1'), Decimal('0.2'))

    def test_convert_selection(self):
        "Test convert selection"
        field = {
            'type': 'selection',
            'selection': [
                ('male', 'Male'),
                ('female', 'Female'),
            ],
        }
        field_with_empty = field.copy()
        field_with_empty['selection'] = (field_with_empty['selection']
                                         + [('', '')])
        for value, result in (
                ('Male', 'male'),
                ('male', 'male'),
                ('test', 'test'),
                (None, None),
                ('', ''),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))
            self.assertEqual(
                convert_value(field_with_empty, value), result,
                msg="convert_value(%r, %r)" % (field_with_empty, value))

    def test_convert_datetime(self):
        "Test convert datetime"
        field = {
            'type': 'datetime',
            'format': '"%H:%M:%S"',
        }
        for value, result in (
                ('12/04/2002', untimezoned_date(dt.datetime(2002, 12, 4))),
                ('12/04/2002 12:30:00', untimezoned_date(
                    dt.datetime(2002, 12, 4, 12, 30))),
                ('02/03/04', untimezoned_date(dt.datetime(2004, 2, 3))),
                ('02/03/04 05:06:07', untimezoned_date(
                    dt.datetime(2004, 2, 3, 5, 6, 7))),
                ('test', None),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_date(self):
        "Test convert date"
        field = {
            'type': 'date',
        }
        for value, result in (
                ('12/04/2002', dt.date(2002, 12, 4)),
                ('test', None),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_time(self):
        "Test convert time"
        field = {
            'type': 'time',
            'format': '"%H:%M:%S"',
        }
        for value, result in (
                ('12:30:00', dt.time(12, 30, 0)),
                ('test', None),
                (None, None),
        ):
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_convert_timedelta(self):
        "Test convert timedelta"
        field = {
            'type': 'timedelta',
        }
        for value, result in [
                ('1d 2:00', dt.timedelta(days=1, hours=2)),
                ('foo', dt.timedelta()),
                (None, None),
        ]:
            self.assertEqual(
                convert_value(field, value), result,
                msg="convert_value(%r, %r)" % (field, value))

    def test_format_boolean(self):
        "Test format boolean"
        field = {
            'type': 'boolean',
        }
        for value, result in (
                (True, 'True'),
                (False, 'False'),
                (None, ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_integer(self):
        "Test format integer"
        field = {
            'type': 'integer',
        }
        for value, result in (
                (1, '1'),
                (1.5, '1'),
                (0, '0'),
                (0.0, '0'),
                (False, ''),
                (None, ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_integer_factor(self):
        "Test format integer with factor"
        field = {
            'type': 'integer',
            'factor': '2',
        }
        self.assertEqual(format_value(field, 3), '6')

    def test_format_float(self):
        "Test format float"
        field = {
            'type': 'float',
        }
        for value, result in (
                (1, '1'),
                (1.5, '1.5'),
                (1.50, '1.5'),
                (150.79, '150.79'),
                (0, '0'),
                (0.0, '0'),
                (False, ''),
                (None, ''),
                (1e-12, '0.000000000001'),
                (1.0579e-10, '0.00000000010579'),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_float_factor(self):
        "Test format float with factor"
        field = {
            'type': 'float',
            'factor': '100',
        }
        self.assertEqual(format_value(field, 0.42), '42')

    def test_format_numeric(self):
        "Test format numeric"
        field = {
            'type': 'numeric',
        }
        for value, result in (
                (Decimal(1), '1'),
                (Decimal('1.5'), '1.5'),
                (Decimal('1.50'), '1.5'),
                (Decimal('150.79'), '150.79'),
                (Decimal(0), '0'),
                (Decimal('0.0'), '0'),
                (False, ''),
                (None, ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_numeric_factor(self):
        "Test format numeric with factor"
        field = {
            'type': 'numeric',
            'factor': '5',
        }
        self.assertEqual(format_value(field, Decimal('0.2')), '1')

    def test_format_selection(self):
        "Test format selection"
        field = {
            'type': 'selection',
            'selection': [
                ('male', 'Male'),
                ('female', 'Female'),
            ],
        }
        field_with_empty = field.copy()
        field_with_empty['selection'] = (field_with_empty['selection']
                                         + [('', '')])
        for value, result in (
                ('male', 'Male'),
                ('test', 'test'),
                (False, ''),
                (None, ''),
                ('', ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))
            self.assertEqual(
                format_value(field_with_empty, value), result,
                msg="format_value(%r, %r)" % (field_with_empty, value))

    def test_format_datetime(self):
        "Test format datetime"
        field = {
            'type': 'datetime',
            'format': '"%H:%M:%S"',
        }
        for value, result in (
                (dt.date(2002, 12, 4), dt.date(2002, 12, 4).strftime('%x')),
                (untimezoned_date(dt.datetime(2002, 12, 4)),
                    dt.date(2002, 12, 4).strftime('%x')),
                (untimezoned_date(dt.datetime(2002, 12, 4, 12, 30)),
                    dt.datetime(2002, 12, 4, 12, 30).strftime(
                        '"%x %H:%M:%S"')),
                (False, ''),
                (None, ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_date(self):
        "Test format date"
        field = {
            'type': 'date',
        }
        for value, result in (
                (dt.date(2002, 12, 4), dt.date(2002, 12, 4).strftime('%x')),
                (False, ''),
                (None, ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_time(self):
        "Test format time"
        field = {
            'type': 'time',
            'format': '"%H:%M:%S"',
        }
        for value, result in (
                (dt.time(12, 30, 0), '"12:30:00"'),
                (False, ''),
                (None, ''),
        ):
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_format_timedelta(self):
        "Test format timedelta"
        field = {
            'type': 'timedelta',
        }
        for value, result in [
                (dt.timedelta(days=1, hours=2), '"1d 02:00"'),
                (dt.timedelta(), ''),
                (None, ''),
                ('', ''),
        ]:
            self.assertEqual(
                format_value(field, value), result,
                msg="format_value(%r, %r)" % (field, value))

    def test_complete_boolean(self):
        "Test complete boolean"
        field = {
            'type': 'boolean',
        }
        for value, result in [
                (None, [True, False]),
                (True, [False]),
                (False, [True]),
        ]:
            self.assertEqual(
                list(complete_value(field, value)), result,
                msg="complete_value(%r, %r)" % (field, value))

    def test_complete_selection(self):
        "Test complete selection"
        field = {
            'type': 'selection',
            'selection': [
                ('male', 'Male'),
                ('female', 'Female'),
            ],
        }
        for value, result in (
                ('m', ['male']),
                ('test', []),
                ('', ['male', 'female']),
                (None, ['male', 'female']),
                (['male', 'f'], [['male', 'female']]),
                (['male', None], [['male', 'male'], ['male', 'female']]),
        ):
            self.assertEqual(
                list(complete_value(field, value)), result,
                msg="complete_value(%r, %r)" % (field, value))

        field_with_empty = field.copy()
        field_with_empty['selection'] = (field_with_empty['selection']
                                         + [('', '')])
        for value, result in (
                ('m', ['male']),
                ('test', []),
                ('', ['male', 'female', '']),
                (None, ['male', 'female', '']),
                (['male', 'f'], [['male', 'female']]),
        ):
            self.assertEqual(
                list(complete_value(field_with_empty, value)), result,
                msg="complete_value(%r, %r)" % (field_with_empty, value))

    def test_complete_reference(self):
        "Test complete reference"
        field = {
            'type': 'reference',
            'selection': [
                ('spam', 'Spam'),
                ('ham', 'Ham'),
                ('', ''),
            ],
        }
        for value, result in (
                ('s', ['%spam%']),
                ('test', []),
                ('', ['%spam%', '%ham%', '%']),
                (None, ['%spam%', '%ham%', '%']),
                (['spam', 'h'], [['spam', 'ham']]),
        ):
            self.assertEqual(
                list(complete_value(field, value)), result,
                msg="complete_value(%r, %r)" % (field, value))

    def test_parenthesize(self):
        for value, result in (
                (['a'], ['a']),
                (['a', 'b'], ['a', 'b']),
                (['(', 'a', ')'], [['a']]),
                (['a', 'b', '(', 'c', '(', 'd', 'e', ')', 'f', ')', 'g'],
                    ['a', 'b', ['c', ['d', 'e'], 'f'], 'g']),
                (['a', 'b', '(', 'c'], ['a', 'b', ['c']]),
                (['a', 'b', '(', 'c', '(', 'd', 'e', ')', 'f'],
                    ['a', 'b', ['c', ['d', 'e'], 'f']]),
                (['a', 'b', ')'], ['a', 'b']),
                (['a', 'b', ')', 'c', ')', 'd)'], ['a', 'b']),
        ):
            self.assertEqual(
                rlist(parenthesize(iter(value))), result,
                msg="parenthesize(%r)" % value)

    def test_operatorize(self):
        "Test operatorize"
        a = ('a', 'a', 'a')
        b = ('b', 'b', 'b')
        c = ('c', 'c', 'c')
        null_ = ('d', None, 'x')
        double_null_ = ('e', None, None)
        for value, result in (
                (['a'], ['a']),
                (['a', '|', 'b'], [['OR', 'a', 'b']]),
                (['a', '|', 'b', '|', 'c'], [['OR', ['OR', 'a', 'b'], 'c']]),
                (['a', 'b', '|', 'c'], ['a', ['OR', 'b', 'c']]),
                (['a', '|', 'b', 'c'], [['OR', 'a', 'b'], 'c']),
                (['a', iter(['b', 'c'])], ['a', ['b', 'c']]),
                (['a', iter(['b', 'c']), 'd'], ['a', ['b', 'c'], 'd']),
                (['a', '|', iter(['b', 'c'])], [['OR', 'a', ['b', 'c']]]),
                (['a', '|', iter(['b', 'c']), 'd'],
                    [['OR', 'a', ['b', 'c']], 'd']),
                (['a', iter(['b', 'c']), '|', 'd'],
                    ['a', ['OR', ['b', 'c'], 'd']]),
                (['a', '|', iter(['b', '|', 'c'])],
                    [['OR', 'a', [['OR', 'b', 'c']]]]),
                (['|'], []),
                (['|', 'a'], ['a']),
                (['a', iter(['|', 'b'])], ['a', ['b']]),
                (['a', '|', '|', 'b'], [['OR', 'a', 'b']]),
                (['|', '|', 'a'], ['a']),
                (['|', '|', 'a', 'b'], ['a', 'b']),
                (['|', '|', 'a', '|', 'b'], [['OR', 'a', 'b']]),
                (['a', iter(['b', '|', 'c'])], ['a', [['OR', 'b', 'c']]]),
                ([a, iter([b, ('|',), c])], [a, [['OR', b, c]]]),
                (['a', iter(['b', '|'])], ['a', [['OR', 'b']]]),
                ([null_], [null_]),
                ([null_, '|', double_null_], [['OR', null_, double_null_]]),
        ):
            self.assertEqual(
                rlist(operatorize(iter(value))), result,
                msg="operatorize(%r)" % value)

    def test_stringable(self):
        "Test stringable"
        dom = DomainParser({
            'name': {
                'string': 'Name',
                'type': 'char',
            },
            'multiselection': {
                'string': "MultiSelection",
                'type': 'multiselection',
                'selection': [
                        ('foo', "Foo"),
                        ('bar', "Bar"),
                        ('baz', "Baz"),
                ],
            },
            'relation': {
                'string': 'Relation',
                'type': 'many2one',
                'relation_fields': {
                        'name': {
                            'string': "Name",
                            'type': 'char',
                        },
                },
            },
            'relations': {
                'string': 'Relations',
                'type': 'many2many',
            },
        })
        valid = ('name', '=', 'Doe')
        invalid = ('surname', '=', 'John')
        self.assertTrue(dom.stringable([]))
        self.assertTrue(dom.stringable([[]]))
        self.assertTrue(dom.stringable([valid]))
        self.assertFalse(dom.stringable([invalid]))
        self.assertTrue(dom.stringable(['AND', valid]))
        self.assertFalse(dom.stringable(['AND', valid, invalid]))
        self.assertTrue(dom.stringable([[valid]]))
        self.assertFalse(dom.stringable([[valid], [invalid]]))
        self.assertTrue(dom.stringable([('multiselection', '=', None)]))
        self.assertTrue(dom.stringable([('multiselection', '=', '')]))
        self.assertFalse(dom.stringable([('multiselection', '=', 'foo')]))
        self.assertTrue(dom.stringable([('multiselection', '=', ['foo'])]))
        self.assertTrue(dom.stringable([('relation', '=', None)]))
        self.assertTrue(dom.stringable([('relation', '=', "Foo")]))
        self.assertTrue(dom.stringable([('relation.rec_name', '=', "Foo")]))
        self.assertFalse(dom.stringable([('relation', '=', 1)]))
        self.assertTrue(dom.stringable([('relations', '=', "Foo")]))
        self.assertTrue(dom.stringable([('relations', '=', None)]))
        self.assertTrue(dom.stringable([('relations', 'in', ["Foo"])]))
        self.assertFalse(dom.stringable([('relations', 'in', [42])]))
        self.assertTrue(dom.stringable([('relation.name', '=', "Foo")]))

    def test_string(self):
        dom = DomainParser({
            'name': {
                'string': 'Name',
                'type': 'char',
            },
            'surname': {
                'string': '(Sur)Name',
                'type': 'char',
            },
            'date': {
                'string': 'Date',
                'type': 'date',
            },
            'selection': {
                'string': 'Selection',
                'type': 'selection',
                'selection': [
                        ('male', 'Male'),
                        ('female', 'Female'),
                        ('', ''),
                ],
            },
            'multiselection': {
                'string': "MultiSelection",
                'type': 'multiselection',
                'selection': [
                        ('foo', "Foo"),
                        ('bar', "Bar"),
                        ('baz', "Baz"),
                ],
            },
            'reference': {
                'string': 'Reference',
                'type': 'reference',
                'selection': [
                        ('spam', 'Spam'),
                        ('ham', 'Ham'),
                ]
            },
            'many2one': {
                'string': 'Many2One',
                'name': 'many2one',
                'type': 'many2one',
                'relation_fields': {
                        'name': {
                            'string': "Name",
                            'type': 'char',
                        },
                },
            },
        })
        self.assertEqual(dom.string([('name', '=', 'Doe')]), 'Name: =Doe')
        self.assertEqual(dom.string([('name', '=', None)]), 'Name: =')
        self.assertEqual(dom.string([('name', '=', '')]), 'Name: =""')
        self.assertEqual(dom.string([('name', 'ilike', '%')]), 'Name: ')
        self.assertEqual(dom.string([('name', 'ilike', '%Doe%')]), 'Name: Doe')
        self.assertEqual(
            dom.string([('name', 'ilike', '%<%')]), 'Name: "" "<"')
        self.assertEqual(dom.string([('name', 'ilike', 'Doe')]), 'Name: =Doe')
        self.assertEqual(dom.string([('name', 'ilike', 'Doe%')]), 'Name: Doe%')
        self.assertEqual(
            dom.string([('name', 'ilike', 'Doe\\%')]), 'Name: =Doe%')
        self.assertEqual(
            dom.string([('name', 'not ilike', '%Doe%')]), 'Name: !Doe')
        self.assertEqual(
            dom.string([('name', 'in', ['John', 'Jane'])]), 'Name: John;Jane')
        self.assertEqual(
            dom.string([('name', 'not in', ['John', 'Jane'])]),
            'Name: !John;Jane')
        self.assertEqual(
            dom.string([
                ('name', 'ilike', '%Doe%'),
                ('name', 'ilike', '%Jane%')]),
            'Name: Doe Name: Jane')
        self.assertEqual(
            dom.string(['AND',
                        ('name', 'ilike', '%Doe%'),
                        ('name', 'ilike', '%Jane%')]),
            'Name: Doe Name: Jane')
        self.assertEqual(
            dom.string(['OR',
                        ('name', 'ilike', '%Doe%'),
                        ('name', 'ilike', '%Jane%')]),
            'Name: Doe | Name: Jane')
        self.assertEqual(
            dom.string([
                ('name', 'ilike', '%Doe%'),
                ['OR',
                    ('name', 'ilike', '%John%'),
                    ('name', 'ilike', '%Jane%')]]),
            'Name: Doe (Name: John | Name: Jane)')
        self.assertEqual(dom.string([]), '')
        self.assertEqual(
            dom.string([('surname', 'ilike', '%Doe%')]), '"(Sur)Name": Doe')
        self.assertEqual(
            dom.string([('date', '>=', dt.date(2012, 10, 24))]),
            dt.date(2012, 10, 24).strftime('Date: >=%x'))
        self.assertEqual(dom.string([('selection', '=', '')]), 'Selection: ')
        self.assertEqual(dom.string([('selection', '=', None)]), 'Selection: ')
        self.assertEqual(
            dom.string([('selection', '!=', '')]), 'Selection: !""')
        self.assertEqual(
            dom.string([('selection', '=', 'male')]), 'Selection: Male')
        self.assertEqual(
            dom.string([('selection', '!=', 'male')]), 'Selection: !Male')
        self.assertEqual(
            dom.string([('multiselection', '=', None)]), "MultiSelection: =")
        self.assertEqual(
            dom.string([('multiselection', '=', '')]), "MultiSelection: =")
        self.assertEqual(
            dom.string([('multiselection', '!=', '')]), "MultiSelection: !=")
        self.assertEqual(
            dom.string([('multiselection', '=', ['foo'])]),
            "MultiSelection: =Foo")
        self.assertEqual(
            dom.string([('multiselection', '!=', ['foo'])]),
            "MultiSelection: !=Foo")
        self.assertEqual(
            dom.string([('multiselection', '=', ['foo', 'bar'])]),
            "MultiSelection: =Foo;Bar")
        self.assertEqual(
            dom.string([('multiselection', '!=', ['foo', 'bar'])]),
            "MultiSelection: !=Foo;Bar")
        self.assertEqual(
            dom.string([('multiselection', 'in', ['foo'])]),
            "MultiSelection: Foo")
        self.assertEqual(
            dom.string([('multiselection', 'not in', ['foo'])]),
            "MultiSelection: !Foo")
        self.assertEqual(
            dom.string([('multiselection', '=', ['foo', 'bar'])]),
            "MultiSelection: =Foo;Bar")
        self.assertEqual(
            dom.string([('multiselection', '!=', ['foo', 'bar'])]),
            "MultiSelection: !=Foo;Bar")
        self.assertEqual(
            dom.string([('multiselection', 'in', ['foo', 'bar'])]),
            "MultiSelection: Foo;Bar")
        self.assertEqual(
            dom.string([('multiselection', 'not in', ['foo', 'bar'])]),
            "MultiSelection: !Foo;Bar")
        self.assertEqual(
            dom.string([('reference', 'ilike', '%foo%')]),
            'Reference: foo')
        self.assertEqual(
            dom.string([('reference.rec_name', 'ilike', '%bar%', 'spam')]),
            'Reference: Spam,bar')
        self.assertEqual(
            dom.string([('reference', 'in', ['foo', 'bar'])]),
            'Reference: foo;bar')
        self.assertEqual(
            dom.string([('many2one', 'ilike', '%John%')]), 'Many2One: John')
        self.assertEqual(
            dom.string([('many2one.rec_name', 'in', ['John', 'Jane'])]),
            'Many2One: John;Jane')
        self.assertEqual(
            dom.string([('many2one.name', 'ilike', '%Foo%')]),
            "Many2One.Name: Foo")

    def test_group(self):
        "Test group"
        dom = DomainParser({
            'name': {
                'string': 'Name',
            },
            'firstname': {
                'string': 'First Name',
            },
            'surname': {
                'string': '(Sur)Name',
            },
            'relation': {
                'string': "Relation",
                'relation': 'relation',
                'relation_fields': {
                    'name': {
                            'string': "Name",
                    },
                },
            },
        })
        self.assertEqual(
            rlist(dom.group(udlex('Name: Doe'))), [('Name', None, 'Doe')])
        self.assertEqual(
            rlist(dom.group(udlex('"(Sur)Name": Doe'))), [
                ('(Sur)Name', None, 'Doe'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: Doe Name: John'))), [
                ('Name', None, 'Doe'),
                ('Name', None, 'John'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: Name: John'))), [
                ('Name', None, None),
                ('Name', None, 'John')])
        self.assertEqual(
            rlist(dom.group(udlex('First Name: John'))), [
                ('First Name', None, 'John'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: Doe First Name: John'))), [
                ('Name', None, 'Doe'),
                ('First Name', None, 'John'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('First Name: John Name: Doe'))), [
                ('First Name', None, 'John'),
                ('Name', None, 'Doe'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('First Name: John First Name: Jane'))), [
                ('First Name', None, 'John'),
                ('First Name', None, 'Jane'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: John Doe'))), [
                ('Name', None, 'John'),
                ('Doe',),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: "John Doe"'))), [
                ('Name', None, 'John Doe'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Doe Name: John'))), [
                ('Doe',),
                ('Name', None, 'John'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: =Doe'))), [('Name', '=', 'Doe')])
        self.assertEqual(
            rlist(dom.group(udlex('Name: =Doe Name: >John'))), [
                ('Name', '=', 'Doe'),
                ('Name', '>', 'John'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('First Name: =John First Name: =Jane'))), [
                ('First Name', '=', 'John'),
                ('First Name', '=', 'Jane'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: John;Jane'))), [
                ('Name', None, ['John', 'Jane'])
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: John;'))), [
                ('Name', None, ['John'])
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: John;Jane Name: Doe'))), [
                ('Name', None, ['John', 'Jane']),
                ('Name', None, 'Doe'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: John; Name: Doe'))), [
                ('Name', None, ['John']),
                ('Name', None, 'Doe'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name:'))), [
                ('Name', None, None),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: ='))), [
                ('Name', '=', None),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: =""'))), [
                ('Name', '=', ''),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: = ""'))), [
                ('Name', '=', ''),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: = Name: Doe'))), [
                ('Name', '=', None),
                ('Name', None, 'Doe'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: \\"foo\\"'))), [
                ('Name', None, '"foo"'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Name: "" <'))), [
                ('Name', '', '<'),
            ])
        self.assertEqual(
            rlist(dom.group(udlex('Relation.Name: Test'))), [
                ('Relation.Name', None, "Test"),
            ])

    def test_parse_clause(self):
        "Test parse clause"
        dom = DomainParser({
            'name': {
                'string': 'Name',
                'name': 'name',
                'type': 'char',
            },
            'integer': {
                'string': 'Integer',
                'name': 'integer',
                'type': 'integer',
            },
            'selection': {
                'string': 'Selection',
                'name': 'selection',
                'type': 'selection',
                'selection': [
                        ('male', 'Male'),
                        ('female', 'Female'),
                ],
            },
            'multiselection': {
                'string': "MultiSelection",
                'name': 'multiselection',
                'type': 'multiselection',
                'selection': [
                        ('foo', "Foo"),
                        ('bar', "Bar"),
                        ('baz', "Baz"),
                ],
            },
            'reference': {
                'string': 'Reference',
                'name': 'reference',
                'type': 'reference',
                'selection': [
                        ('spam', 'Spam'),
                        ('ham', 'Ham'),
                ]
            },
            'many2one': {
                'string': 'Many2One',
                'name': 'many2one',
                'type': 'many2one',
            },
            'relation': {
                'string': "Relation",
                'relation': 'relation',
                'name': 'relation',
                'relation_fields': {
                        'name': {
                            'string': "Name",
                            'name': 'name',
                            'type': 'char',
                        },
                },
            },
        })
        self.assertEqual(
            rlist(dom.parse_clause([('John',)])), [
                ('rec_name', 'ilike', '%John%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', None, None)])), [
                ('name', 'ilike', '%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', '', None)])), [
                ('name', 'ilike', '%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', '=', None)])), [
                ('name', '=', None),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', '=', '')])), [
                ('name', '=', ''),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', None, 'Doe')])), [
                ('name', 'ilike', '%Doe%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', '!', 'Doe')])), [
                ('name', 'not ilike', '%Doe%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', None, ['John', 'Jane'])])), [
                ('name', 'in', ['John', 'Jane']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Name', '!', ['John', 'Jane'])])), [
                ('name', 'not in', ['John', 'Jane']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Selection', None, None)])), [
                ('selection', '=', None),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Selection', None, '')])), [
                ('selection', '=', ''),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Selection', None, ['Male', 'Female'])])),
            [('selection', 'in', ['male', 'female'])])
        self.assertEqual(
            rlist(dom.parse_clause([('MultiSelection', None, None)])), [
                ('multiselection', '=', None),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('MultiSelection', None, '')])), [
                ('multiselection', 'in', ['']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('MultiSelection', '=', '')])), [
                ('multiselection', '=', ['']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('MultiSelection', '!', '')])), [
                ('multiselection', 'not in', ['']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('MultiSelection', '!=', '')])), [
                ('multiselection', '!=', ['']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause(
                [('MultiSelection', None, ['Foo', 'Bar'])])), [
                ('multiselection', 'in', ['foo', 'bar']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause(
                [('MultiSelection', '=', ['Foo', 'Bar'])])), [
                ('multiselection', '=', ['foo', 'bar']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause(
                [('MultiSelection', '!', ['Foo', 'Bar'])])), [
                ('multiselection', 'not in', ['foo', 'bar']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause(
                [('MultiSelection', '!=', ['Foo', 'Bar'])])), [
                ('multiselection', '!=', ['foo', 'bar']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Integer', None, None)])), [
                ('integer', '=', None),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Integer', None, '3..5')])), [[
                ('integer', '>=', 3),
                ('integer', '<=', 5),
            ]])
        self.assertEqual(
            rlist(dom.parse_clause([('Reference', None, 'foo')])), [
                ('reference', 'ilike', '%foo%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Reference', None, 'Spam')])), [
                ('reference', 'ilike', '%spam%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Reference', None, 'Spam,bar')])), [
                ('reference.rec_name', 'ilike', '%bar%', 'spam'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Reference', None, ['foo', 'bar'])])), [
                ('reference', 'in', ['foo', 'bar']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause(['OR',
                                    ('Name', None, 'John'),
                                    ('Name', None, 'Jane')])),
            ['OR',
                ('name', 'ilike', '%John%'),
                ('name', 'ilike', '%Jane%'),
             ])
        self.assertEqual(
            rlist(dom.parse_clause([('Many2One', None, 'John')])), [
                ('many2one', 'ilike', '%John%'),
            ])
        self.assertEqual(
            rlist(dom.parse_clause([('Many2One', None, ['John', 'Jane'])])), [
                ('many2one.rec_name', 'in', ['John', 'Jane']),
            ])
        self.assertEqual(
            rlist(dom.parse_clause(iter([iter([['John']])]))), [
                [('rec_name', 'ilike', '%John%')]])
        self.assertEqual(
            rlist(dom.parse_clause(iter([['Relation.Name', None, "Test"]]))),
            [('relation.name', 'ilike', "%Test%")])
        self.assertEqual(
            rlist(dom.parse_clause(iter([['OR']]))),
            [('rec_name', 'ilike', "%OR%")])
        self.assertEqual(
            rlist(dom.parse_clause(iter([['AND']]))),
            [('rec_name', 'ilike', "%AND%")])

    def test_completion_char(self):
        "Test completion char"
        dom = DomainParser({
            'name': {
                'string': 'Name',
                'name': 'name',
                'type': 'char',
            },
        })
        self.assertEqual(list(dom.completion('Nam')), ['Name: '])
        self.assertEqual(list(dom.completion('Name:')), ['Name: '])
        self.assertEqual(list(dom.completion('Name: foo')), [])
        self.assertEqual(list(dom.completion('Name: !=')), [])
        self.assertEqual(list(dom.completion('Name: !=foo')), [])
        self.assertEqual(list(dom.completion('')), ['Name: '])
        self.assertEqual(list(dom.completion(' ')), ['', 'Name: '])
        self.assertEqual(list(dom.completion('Name: foo |')), ['Name: foo'])
        self.assertEqual(
            list(dom.completion('Name: foo (Name: foo | N')),
            ['Name: foo (Name: foo | Name: '])

    def test_completion_many2one(self):
        "Test completion many2one"
        dom = DomainParser({
            'relation': {
                'name': 'relation',
                'string': "Relation",
                'type': 'many2one',
                'relation_fields': {
                        'name': {
                            'name': 'name',
                            'string': "Name",
                            'type': 'char',
                        },
                },
            },
        })
        self.assertEqual(
            list(dom.completion('Relatio')),
            ['Relation: ', 'Relation.Name: '])

    def test_completion_boolean(self):
        "Test completion boolean"
        dom = DomainParser({
            'name': {
                'string': "Active",
                'name': 'active',
                'type': 'boolean',
            },
        })

        self.assertEqual(list(dom.completion("Act")), ["Active: "])
        self.assertEqual(list(dom.completion("Active:")),
                         ["Active: ", "Active: True", "Active: False"])
        self.assertEqual(
            list(dom.completion("Active: t")),
            ["Active: True", "Active: False"])
        self.assertEqual(
            list(dom.completion("Active: f")),
            ["Active: False", "Active: True"])



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
"Board"
import gettext
import xml.dom.minidom

from gnuhealth.common import MODELNAME, RPCExecute
from gnuhealth.gui import Main
from gnuhealth.gui.window.view_board import ViewBoard

from .tabcontent import TabContent

_ = gettext.gettext


class Board(TabContent):
    'Board'

    def __init__(self, model, name='', **attributes):
        super(Board, self).__init__(**attributes)

        context = attributes.get('context')
        self.view_id, = attributes.get('view_ids')

        view = RPCExecute(
            'model', 'ir.ui.view', 'view_get', self.view_id, context=context)

        xml_dom = xml.dom.minidom.parseString(view['arch'])
        root, = xml_dom.childNodes
        self.board = ViewBoard(root, context=context)
        self.model = model
        self.dialogs = []
        if not name:
            name = MODELNAME.get(model)
        self.name = name

        self.create_tabcontent()
        self.board.reload()

    def get_toolbars(self):
        return {}

    def widget_get(self):
        return self.board.widget_get()

    def sig_reload(self, test_modified=True):
        self.board.reload()
        return True

    def sig_close(self):
        return True

    def compare(self, model, attributes):
        if not attributes:
            return False
        return (
            self.model == model
            and self.attributes.get('view_ids') == attributes.get('view_ids')
            and self.attributes.get('context') == attributes.get('context'))

    def __hash__(self):
        return id(self)

    def sig_win_close(self, widget):
        Main().sig_win_close(widget)

    def set_cursor(self):
        if not self.board.actions:
            return
        first_action = self.board.actions[0]
        first_action.screen.set_cursor()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
'Action'
import gettext

from gi.repository import GLib, Gtk

import gnuhealth.rpc as rpc
from gnuhealth.action import Action as GenericAction
from gnuhealth.common import RPCException, RPCExecute
from gnuhealth.config import CONFIG
from gnuhealth.gui.window.view_form.screen import Screen
from gnuhealth.gui.window.win_form import WinForm
from gnuhealth.pyson import PYSONDecoder

_ = gettext.gettext


class Action:

    def __init__(self, view, attrs=None):
        super(Action, self).__init__()
        self.name = attrs['name']
        self.view = view

        try:
            self.action = RPCExecute('model', 'ir.action.act_window', 'get',
                                     self.name)
        except RPCException:
            raise

        view_ids = []
        view_mode = None
        if self.action.get('views', []):
            view_ids = [x[0] for x in self.action['views']]
            view_mode = [x[1] for x in self.action['views']]
        elif self.action.get('view_id', False):
            view_ids = [self.action['view_id'][0]]

        self.action.setdefault('pyson_domain', '[]')
        ctx = {}
        ctx.update(rpc.CONTEXT)
        ctx['_user'] = rpc._USER
        decoder = PYSONDecoder(ctx)
        action_ctx = self.view.context.copy()
        action_ctx.update(
            decoder.decode(self.action.get('pyson_context') or '{}'))
        ctx.update(action_ctx)

        ctx['context'] = ctx
        decoder = PYSONDecoder(ctx)
        order = decoder.decode(self.action['pyson_order'])
        search_value = decoder.decode(
            self.action['pyson_search_value'] or '[]')
        tab_domain = [(n, decoder.decode(d), c)
                      for n, d, c in self.action['domains']]

        limit = self.action.get('limit')
        if limit is None:
            limit = CONFIG['client.limit']

        self.context = ctx
        self.domain = []
        self.update_domain([])

        self.widget = Gtk.Frame()
        self.widget.set_border_width(0)

        vbox = Gtk.VBox(homogeneous=False, spacing=3)
        self.widget.add(vbox)

        self.title = Gtk.Label()
        self.widget.set_label_widget(self.title)
        self.widget.set_label_align(0.0, 0.5)
        self.widget.show_all()

        self.screen = Screen(self.action['res_model'],
                             view_ids=view_ids,
                             domain=self.domain,
                             context=action_ctx,
                             order=order,
                             mode=view_mode,
                             limit=limit,
                             search_value=search_value,
                             tab_domain=tab_domain,
                             context_model=self.action['context_model'],
                             context_domain=self.action['context_domain'],
                             row_activate=self.row_activate)
        self.screen.windows.append(self)
        vbox.pack_start(
            self.screen.widget, expand=True, fill=True, padding=0)

        if attrs.get('string'):
            self.title.set_text(attrs['string'])
        else:
            self.title.set_text(self.action['name'])

        self.widget.set_size_request(int(attrs.get('width', -1)),
                                     int(attrs.get('height', -1)))

        self.screen.search_filter()

    def row_activate(self):
        if not self.screen.current_record:
            return

        if (self.screen.current_view.view_type == 'tree'
                and int(self.screen.current_view.attributes.get(
                        'keyword_open', 0))):
            GenericAction.exec_keyword('tree_open', {
                'model': self.screen.model_name,
                'id': (self.screen.current_record.id
                       if self.screen.current_record else None),
                'ids': [r.id for r in self.screen.selected_records],
            }, context=self.screen.group._context.copy(),
                warning=False)
        else:
            def callback(result):
                if result:
                    self.screen.current_record.save()
                else:
                    self.screen.current_record.cancel()
            WinForm(self.screen, callback, title=self.title.get_text())

    def display(self):
        self.screen.search_filter(self.screen.screen_container.get_text())

    def record_message(self, *args):
        self.view.active_changed(self)

    def _get_active(self):
        if self.screen and self.screen.current_record:
            return {
                'active_id': self.screen.current_record.id,
                'active_ids': [r.id for r in self.screen.selected_records]
            }
        else:
            return {
                'active_id': None,
                'active_ids': [],
            }

    active = property(_get_active)

    def update_domain(self, actions):
        domain_ctx = self.context.copy()
        domain_ctx['_actions'] = {}
        for action in actions:
            domain_ctx['_actions'][action.name] = action.active
        new_domain = PYSONDecoder(domain_ctx).decode(
            self.action['pyson_domain'])
        if self.domain == new_domain:
            return
        del self.domain[:]
        self.domain.extend(new_domain)
        if hasattr(self, 'screen'):  # Catch early update
            # Using idle_add to prevent corruption of the event who triggered
            # the update.
            def display():
                if self.screen.widget.props.window:
                    self.display()
            GLib.idle_add(display)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gi.repository import Gdk, Gtk

import gnuhealth.common as common
from gnuhealth.common.underline import set_underline
from gnuhealth.config import GNUHEALTH_ICON
from gnuhealth.gui import Main
from gnuhealth.gui.window.nomodal import NoModal
from gnuhealth.gui.window.view_form.screen import Screen
from gnuhealth.gui.window.win_form import WinForm

_ = gettext.gettext


class WinSearch(NoModal):

    def __init__(self, model, callback, sel_multi=True, context=None,
                 domain=None, order=None, view_ids=None,
                 views_preload=None, new=True, title='', exclude_field=None):
        NoModal.__init__(self)
        if view_ids is None:
            view_ids = []
        if views_preload is None:
            views_preload = {}
        self.domain = domain or []
        self.context = context or {}
        self.order = order
        self.view_ids = view_ids
        self.views_preload = views_preload
        self.sel_multi = sel_multi
        self.callback = callback
        if not title:
            title = common.MODELNAME.get(model)
        self.title = title
        self.exclude_field = exclude_field

        self.win = Gtk.Dialog(
            title=_('Search'), transient_for=self.parent,
            destroy_with_parent=True)
        Main().add_window(self.win)
        self.win.set_icon(GNUHEALTH_ICON)
        self.win.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.win.set_default_response(Gtk.ResponseType.APPLY)
        self.win.connect('response', self.response)

        self.win.set_default_size(*self.default_size())

        self.accel_group = Gtk.AccelGroup()
        self.win.add_accel_group(self.accel_group)

        self.but_cancel = self.win.add_button(
            set_underline(_("Cancel")), Gtk.ResponseType.CANCEL)
        self.but_cancel.set_image(common.IconFactory.get_image(
            'gnuhealth-cancel', Gtk.IconSize.BUTTON))
        self.but_cancel.set_always_show_image(True)
        self.but_find = self.win.add_button(
            set_underline(_("Search")), Gtk.ResponseType.APPLY)
        self.but_find.set_image(common.IconFactory.get_image(
            'gnuhealth-search', Gtk.IconSize.BUTTON))
        self.but_find.set_always_show_image(True)
        if new and common.MODELACCESS[model]['create']:
            self.but_new = self.win.add_button(
                set_underline(_("New")), Gtk.ResponseType.ACCEPT)
            self.but_new.set_image(common.IconFactory.get_image(
                'gnuhealth-create', Gtk.IconSize.BUTTON))
            self.but_new.set_always_show_image(True)
            self.but_new.set_accel_path(
                '<from gnuhealth>/Form/New', self.accel_group)

        self.but_ok = self.win.add_button(
            set_underline(_("OK")), Gtk.ResponseType.OK)
        self.but_ok.set_image(common.IconFactory.get_image(
            'gnuhealth-ok', Gtk.IconSize.BUTTON))
        self.but_ok.set_always_show_image(True)
        self.but_ok.add_accelerator(
            'clicked', self.accel_group, Gdk.KEY_Return,
            Gdk.ModifierType.CONTROL_MASK, Gtk.AccelFlags.VISIBLE)

        hbox = Gtk.HBox()
        hbox.show()
        self.win.vbox.pack_start(hbox, expand=False, fill=True, padding=0)
        self.win.vbox.pack_start(
            Gtk.HSeparator(), expand=False, fill=True, padding=0)

        self.screen = Screen(
            model, domain=domain, mode=['tree'], order=order,
            context=context, view_ids=view_ids, views_preload=views_preload,
            row_activate=self.sig_activate, readonly=True)
        self.view = self.screen.current_view
        # Prevent to set tree_state
        self.screen.tree_states_done.add(id(self.view))
        sel = self.view.treeview.get_selection()
        self.win.set_title(_('Search %s') % self.title)

        if not sel_multi:
            sel.set_mode(Gtk.SelectionMode.SINGLE)
        else:
            sel.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.win.vbox.pack_start(
            self.screen.widget, expand=True, fill=True, padding=0)
        self.screen.widget.show()

        self.model_name = model

        self.register()

    def sig_activate(self, *args):
        self.view.treeview.stop_emission_by_name('row_activated')
        self.win.response(Gtk.ResponseType.OK)
        return True

    def destroy(self):
        self.screen.destroy()
        self.win.destroy()
        NoModal.destroy(self)

    def show(self):
        self.win.show()

    def hide(self):
        self.win.hide()

    def response(self, win, response_id):
        res = None
        if response_id == Gtk.ResponseType.OK:
            res = [r.id for r in self.screen.selected_records]
        elif response_id == Gtk.ResponseType.APPLY:
            self.screen.search_filter(self.screen.screen_container.get_text())
            return
        elif response_id == Gtk.ResponseType.ACCEPT:
            # Remove first tree view as mode if form only
            view_ids = self.view_ids[1:]
            screen = Screen(
                self.model_name, domain=self.domain,
                context=self.context, order=self.order, mode=['form'],
                view_ids=view_ids, views_preload=self.views_preload,
                exclude_field=self.exclude_field,
                breadcrumb=[self.title])

            def callback(result):
                if result:
                    record = screen.current_record
                    res = [(record.id, record.value.get('rec_name', ''))]
                    self.callback(res)
                else:
                    self.callback(None)
            self.destroy()
            WinForm(screen, callback, new=True, save_current=True)
            return
        if res:
            group = self.screen.group
            res = [(id_, group.get(id_).value.get('rec_name', ''))
                   for id_ in res]
        self.callback(res)
        self.destroy()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from .integer import Integer


class Float(Integer):
    "Float"

    @property
    def digits(self):
        if self.field and self.record:
            return self.field.digits(self.record, factor=self.factor)

    @property
    def width(self):
        digits = self.digits
        if digits:
            return sum(digits)
        else:
            return self.attrs.get('width', 18)

    def display(self):
        digits = self.digits
        if digits:
            self.entry.digits = digits[1]
        else:
            self.entry.digits = None
        super(Float, self).display()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

import gettext

try:
    from http import HTTPStatus
except ImportError:
    from http import client as HTTPStatus

from gi.repository import Gtk

from gnuhealth.common import (
    IconFactory, RPCException, play_sound, process_exception)
from gnuhealth.common.underline import set_underline
from gnuhealth.config import CONFIG, GNUHEALTH_ICON
from gnuhealth.exceptions import GNUHealthServerError
from gnuhealth.gui import Main
from gnuhealth.gui.window.nomodal import NoModal

_ = gettext.gettext


class CodeScanner(NoModal):

    def __init__(self, callback, loop=False):
        super().__init__()
        self.callback = callback
        self.loop = loop
        self.dialog = Gtk.MessageDialog(
            transient_for=self.parent, destroy_with_parent=True,
            text=_("Code Scanner"))
        Main().add_window(self.dialog)
        self.dialog.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
        self.dialog.set_icon(GNUHEALTH_ICON)
        self.dialog.connect('response', self.response)

        self.dialog.set_title(_("Code Scanner"))

        self.entry = Gtk.Entry()
        self.entry.set_activates_default(True)
        self.entry.set_placeholder_text(_("Code"))
        self.dialog.get_message_area().pack_start(
            self.entry, expand=False, fill=False, padding=9)

        button_close = self.dialog.add_button(
            set_underline(_("Close")), Gtk.ResponseType.CLOSE)
        button_close.set_image(IconFactory.get_image(
            'gnuhealth-close', Gtk.IconSize.BUTTON))

        button_ok = self.dialog.add_button(
            set_underline(_("OK")), Gtk.ResponseType.OK)
        button_ok.set_image(IconFactory.get_image(
            'gnuhealth-ok', Gtk.IconSize.BUTTON))
        self.dialog.set_default_response(Gtk.ResponseType.OK)

        self.dialog.show_all()
        self.register()
        self.entry.grab_focus()

    def _play(self, sound):
        if CONFIG['client.code_scanner_sound']:
            play_sound(sound)

    def response(self, dialog, response):
        if response == Gtk.ResponseType.OK:
            code = self.entry.get_text()
            self.entry.set_text('')
            if code:
                while True:
                    try:
                        modified = self.callback(code)
                        self._play('success')
                        if not self.loop or not modified:
                            self.destroy()
                    except Exception as exception:
                        unauthorized = (
                            isinstance(exception, GNUHealthServerError)
                            and exception.faultCode == str(
                                int(HTTPStatus.UNAUTHORIZED)))
                        if not unauthorized:
                            self._play('danger')
                        try:
                            process_exception(exception)
                        except RPCException:
                            pass
                        if unauthorized:
                            continue
                        self.destroy()
                    return
        if not self.loop or response != Gtk.ResponseType.OK:
            self.destroy()

    def destroy(self):
        super().destroy()
        self.dialog.destroy()

    def show(self):
        self.dialog.show()

    def hide(self):
        self.dialog.hide()



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import importlib
import os

from gnuhealth.config import CURRENT_DIR, get_config_dir

__all__ = ['MODULES', 'register']

_ = gettext.gettext

MODULES = []


def register():
    global MODULES
    paths = [
        os.path.join(get_config_dir(), 'plugins'),
        os.path.join(CURRENT_DIR, 'plugins'),
        ]
    paths = list(filter(os.path.isdir, paths))

    imported = set()
    for path in paths:
        finder = importlib.machinery.FileFinder(
            path,
            (importlib.machinery.SourceFileLoader,
                importlib.machinery.SOURCE_SUFFIXES),
            (importlib.machinery.SourcelessFileLoader,
                importlib.machinery.BYTECODE_SUFFIXES))
        for plugin in os.listdir(path):
            module = os.path.splitext(plugin)[0]
            if (module.startswith('_') or module in imported):
                continue
            module = 'gnuhealth.plugins.%s' % module
            spec = finder.find_spec(module)
            if not spec:
                continue
            module = importlib.util.module_from_spec(spec)
            try:
                spec.loader.exec_module(module)
            except ImportError:
                continue
            else:
                MODULES.append(module)
                imported.add(module.__name__)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import base64
import json
import logging
import socket
import threading
import time
import uuid
from urllib.error import HTTPError
from urllib.request import Request, urlopen

from gi.repository import GLib

from gnuhealth.config import CONFIG
from gnuhealth.jsonrpc import object_hook

logger = logging.getLogger(__name__)


ID = str(uuid.uuid4())
CHANNELS = [
    'client:%s' % ID,
]


def listen(connection):
    if not CONFIG['thread']:
        return
    listener = threading.Thread(
        target=_listen, args=(connection,), daemon=True)
    listener.start()


def _listen(connection):
    bus_timeout = CONFIG['client.bus_timeout']
    session = connection.session
    authorization = base64.b64encode(session.encode('utf-8'))
    headers = {
        'Content-Type': 'application/json',
        'Authorization': b'Session ' + authorization,
    }

    wait = 1
    last_message = None
    url = None
    while connection.session == session:
        if url is None:
            if connection.url is None:
                time.sleep(1)
                continue
            url = connection.url + '/bus'
        request = Request(url,
                          data=json.dumps({
                              'last_message': last_message,
                              'channels': CHANNELS,
                          }).encode('utf-8'),
                          headers=headers)
        logger.info('poll channels %s with last message %s',
                    CHANNELS, last_message)
        try:
            response = urlopen(request, timeout=bus_timeout)
            wait = 1
        except socket.timeout:
            wait = 1
            continue
        except Exception as error:
            if isinstance(error, HTTPError):
                if error.code in (301, 302, 303, 307, 308):
                    url = error.headers.get('Location')
                    continue
                elif error.code == 501:
                    logger.info("Bus not supported")
                    break
            logger.error(
                "An exception occurred while connecting to the bus. "
                "Sleeping for %s seconds",
                wait, exc_info=error)
            time.sleep(min(wait, bus_timeout))
            wait *= 2
            continue

        if connection.session != session:
            break

        data = json.loads(response.read(), object_hook=object_hook)
        if data['message']:
            last_message = data['message']['message_id']
            GLib.idle_add(handle, data['message'])


def handle(message):
    from gnuhealth.gui.main import Main

    app = Main()
    if message['type'] == 'notification':
        app.show_notification(
            message.get('title', ''), message.get('body', ''),
            message.get('priority', 1))



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext

from gi.repository import Gdk, Gtk

from gnuhealth.action import Action
from gnuhealth.common import RPCException, RPCExecute
from gnuhealth.common.common import selection
from gnuhealth.gui.window import Window
from gnuhealth.gui.window.attachment import Attachment
from gnuhealth.gui.window.email_ import Email
from gnuhealth.gui.window.log import Log
from gnuhealth.gui.window.note import Note
from gnuhealth.gui.window.view_form.screen import Screen

_ = gettext.gettext


def populate(menu, model, record, title='', field=None, context=None):
    '''
    Fill menu with the actions of model for the record.
    If title is filled, the actions will be put in a submenu.
    '''
    if record is None:
        return
    elif isinstance(record, int):
        if record < 0:
            return
    elif record.id < 0:
        return

    def load(record):
        if isinstance(record, int):
            screen = Screen(model, context=context)
            screen.load([record])
            record = screen.group.get(record)
        return record

    def id_(record):
        if not isinstance(record, int):
            return record.id
        return record

    def activate(menuitem, action, atype):
        rec = load(record)
        data = {
            'model': model,
            'id': rec.id,
            'ids': [rec.id],
        }
        event = Gtk.get_current_event()
        allow_similar = False
        if (event.state & Gdk.ModifierType.CONTROL_MASK
                or event.state & Gdk.ModifierType.MOD1_MASK):
            allow_similar = True
        with Window(hide_current=True, allow_similar=allow_similar):
            Action.execute(action, data, context=rec.get_context())

    def log(menuitem):
        Log(load(record))

    def attachment(menuitem):
        Attachment(load(record), None)

    def note(menuitem):
        Note(load(record), None)

    def is_report(action):
        return action['type'] == 'ir.action.report'

    def email(menuitem, toolbar):
        rec = load(record)
        prints = filter(is_report, toolbar['print'])
        emails = {e['name']: e['id'] for e in toolbar['emails']}
        template = selection(_("Template"), emails, alwaysask=True)
        if template:
            template = template[1]
        Email(
            '%s: %s' % (title, rec.rec_name()), rec, prints,
            template=template)

    def edit(menuitem):
        with Window(hide_current=True, allow_similar=True):
            Window.create(model,
                          view_ids=field.attrs.get('view_ids', '').split(','),
                          res_id=id_(record),
                          mode=['form'],
                          name=field.attrs.get('string'),
                          context=context)

    if title:
        if len(menu):
            menu.append(Gtk.SeparatorMenuItem())
        title_item = Gtk.MenuItem(label=title)
        menu.append(title_item)
        submenu = Gtk.Menu()
        title_item.set_submenu(submenu)
        action_menu = submenu
    else:
        action_menu = menu

    if len(action_menu):
        action_menu.append(Gtk.SeparatorMenuItem())
    if field:
        edit_item = Gtk.MenuItem(label=_('Edit...'))
        edit_item.connect('activate', edit)
        action_menu.append(edit_item)
        action_menu.append(Gtk.SeparatorMenuItem())
    log_item = Gtk.MenuItem(label=_("View Logs..."))
    action_menu.append(log_item)
    log_item.connect('activate', log)
    attachment_item = Gtk.MenuItem(label=_('Attachments...'))
    action_menu.append(attachment_item)
    attachment_item.connect('activate', attachment)
    note_item = Gtk.MenuItem(label=_('Notes...'))
    action_menu.append(note_item)
    note_item.connect('activate', note)

    try:
        toolbar = RPCExecute('model', model, 'view_toolbar_get')
    except RPCException:
        pass
    else:
        for atype, icon, label, flavor in (
                ('action', 'gnuhealth-launch', _('Actions...'), None),
                ('relate', 'gnuhealth-link', _('Relate...'), None),
                ('print', 'gnuhealth-open', _('Report...'), 'open'),
                ('print', 'gnuhealth-print', _('Print...'), 'print'),
        ):
            if len(action_menu):
                action_menu.append(Gtk.SeparatorMenuItem())
            title_item = Gtk.MenuItem(label=label)
            action_menu.append(title_item)
            if not toolbar[atype]:
                title_item.set_sensitive(False)
                continue
            submenu = Gtk.Menu()
            title_item.set_submenu(submenu)
            for action in toolbar[atype]:
                action = action.copy()
                item = Gtk.MenuItem(label=action['name'])
                submenu.append(item)
                if flavor == 'print':
                    action['direct_print'] = True
                item.connect('activate', activate, action, atype)
            menu.show_all()

        email_item = Gtk.MenuItem(label=_('E-Mail...'))
        action_menu.append(email_item)
        email_item.connect('activate', email, toolbar)
    menu.show_all()


def popup(menu, widget):
    def menu_position(menu, x, y, user_data):
        widget_allocation = widget.get_allocation()
        x, y = widget.get_window().get_root_coords(
            widget_allocation.x, widget_allocation.y)
        return (x, y + widget_allocation.height, False)
    menu.show_all()
    if hasattr(menu, 'popup_at_widget'):
        menu.popup_at_widget(
            widget, Gdk.Gravity.SOUTH_WEST, Gdk.Gravity.NORTH_WEST,
            Gtk.get_current_event())
    else:
        menu.popup(
            None, None, menu_position, None, 0,
            Gtk.get_current_event_time())



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# SPDX-FileCopyrightText: 2013 Simon Feltman
# SPDX-FileCopyrightText: 2017-2024 Luis Falcón <falcon@gnuhealth.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Solidario <health@gnusolidario.org>

# SPDX-License-Identifier: GPL-3.0-or-later

#########################################################################
#             GNUHEALTH HOSPITAL MANAGEMENT - GTK CLIENT                #
#                     https://www.gnuhealth.org                         #
#########################################################################
#                        generictreemodel.py:                           #
#     GenericTreeModel implementation for pygtk compatibility.          #
#########################################################################

# System
import sys
import random
import collections
import ctypes
import platform

# GObject
from gi.repository import GObject
from gi.repository import Gtk


class _CTreeIter(ctypes.Structure):
    _fields_ = [('stamp', ctypes.c_int),
                ('user_data', ctypes.c_void_p),
                ('user_data2', ctypes.c_void_p),
                ('user_data3', ctypes.c_void_p)]

    @classmethod
    def from_iter(cls, iter):
        offset = sys.getsizeof(object())  # size of PyObject_HEAD
        return ctypes.POINTER(cls).from_address(id(iter) + offset)


if platform.python_implementation() == "PyPy":
    def _get_user_data_as_pyobject(iter):
        raise NotImplementedError("Not yet supported under PyPy")
else:
    def _get_user_data_as_pyobject(iter):
        citer = _CTreeIter.from_iter(iter)
        return ctypes.cast(citer.contents.user_data, ctypes.py_object).value


def handle_exception(default_return):
    """Returns a function which can act as a decorator for wrapping
    exceptions and eturning "default_return" upon an exception being thrown.

    This is used to wrap Gtk.TreeModel "do_" method implementations
    so we can return a proper value from the override upon an exception
    occurring with client code implemented by the "on_" methods.
    """
    def decorator(func):
        def wrapped_func(*args, **kargs):
            try:
                return func(*args, **kargs)
            except Exception:
                # Use excepthook directly to avoid any printing to the screen
                # if someone installed an except hook.
                sys.excepthook(*sys.exc_info())
            return default_return
        return wrapped_func
    return decorator


class GenericTreeModel(GObject.GObject, Gtk.TreeModel):
    """A base implementation of a Gtk.TreeModel for python.

    The GenericTreeModel eases implementing the Gtk.TreeModel
    interface in Python. The class can be subclassed to provide a
    TreeModel implementation which works directly with Python
    objects instead of iterators.

    All of the on_* methods should be overridden by subclasses
    to provide the underlying implementation a way to access custom
    model data. For the purposes of this API, all custom model data
    supplied or handed back through the overridable API will use the
    argument names: node, parent, and child in regards to user data
    python objects.

    The create_tree_iter, set_user_data, invalidate_iters,
    iter_is_valid methods are available to help manage Gtk.TreeIter
    objects and their Python object references.

    GenericTreeModel manages a pool of user data nodes that have been
    used with iters. This pool stores a references to user data nodes
    as a dictionary value with the key being the integer id of the data.
    This id is what the Gtk.TreeIter objects use to reference data in the pool.
    References will be removed from the pool when the model is deleted
    or explicitly by using the optional "node" argument to the "row_deleted"
    method when notifying the model of row deletion.
    """

    leak_references = GObject.Property(
        default=True, type=bool,
        blurb="If True, strong references to user data attached to iters are "
        "stored in a dictionary pool (default). Otherwise the user data is "
        "stored as a raw pointer to a python object without a reference.")

    #
    # Methods
    #
    def __init__(self):
        """Initialize. Make sure to call this from derived classes if
        overridden."""
        super(GenericTreeModel, self).__init__()
        self.stamp = 0

        #: Dictionary of (id(user_data): user_data),
        #  used when leak-refernces=False
        self._held_refs = dict()

        # Set initial stamp
        self.invalidate_iters()

    def iter_depth_first(self):
        """Depth-first iteration of the entire TreeModel yielding
        the python nodes."""
        stack = collections.deque([None])
        while stack:
            it = stack.popleft()
            if it is not None:
                yield self.get_user_data(it)
            children = [
                self.iter_nth_child(it, i)
                for i in range(self.iter_n_children(it))]
            stack.extendleft(reversed(children))

    def invalidate_iter(self, iter):
        """Clear user data and its reference from the iter and this model."""
        iter.stamp = 0
        if iter.user_data:
            if iter.user_data in self._held_refs:
                del self._held_refs[iter.user_data]
            iter.user_data = None

    def invalidate_iters(self):
        """
        This method invalidates all TreeIter objects associated with
        this custom tree model and frees their locally pooled references.
        """
        self.stamp = random.randint(-2147483648, 2147483647)
        self._held_refs.clear()

    def iter_is_valid(self, iter):
        """
        :Returns:
            True if the gtk.TreeIter specified by iter is valid
            for the custom tree model.
        """
        return iter.stamp == self.stamp

    def get_user_data(self, iter):
        """Get the user_data associated with the given TreeIter.

        GenericTreeModel stores arbitrary Python objects mapped
        to instances of Gtk.TreeIter. This method allows to retrieve the
        Python object held by the given iterator.
        """
        if self.leak_references:
            return self._held_refs[iter.user_data]
        else:
            return _get_user_data_as_pyobject(iter)

    def set_user_data(self, iter, user_data):
        """Applies user_data and stamp to the given iter.

        If the models "leak_references" property is set, a reference to the
        user_data is stored with the model to ensure we don't run into bad
        memory problems with the TreeIter.
        """
        iter.user_data = id(user_data)

        if user_data is None:
            self.invalidate_iter(iter)
        else:
            iter.stamp = self.stamp
            if self.leak_references:
                self._held_refs[iter.user_data] = user_data

    def create_tree_iter(self, user_data):
        """Create a Gtk.TreeIter instance with the given user
           data specific for this model.

        Use this method to create Gtk.TreeIter instance
        instead of directly calling Gtk.Treeiter(), this will ensure
        proper reference managment of wrapped used_data.
        """
        iter = Gtk.TreeIter()
        self.set_user_data(iter, user_data)
        return iter

    def _create_tree_iter(self, data):
        """Internal creation of a (bool, TreeIter) pair for returning directly
        back to the view interfacing with this model."""
        if data is None:
            return (False, None)
        else:
            it = self.create_tree_iter(data)
            return (True, it)

    def row_deleted(self, path, node=None):
        """Notify the model a row has been deleted.

        Use the node parameter to ensure the user_data reference associated
        with the path is properly freed by this model.

        :Parameters:
            path : Gtk.TreePath
                Path to the row that has been deleted.
            node : object
                Python object used as the node returned from
                "on_get_iter". This is optional but ensures the model
                will not leak references to this object.
        """
        super(GenericTreeModel, self).row_deleted(path)
        node_id = id(node)
        if node_id in self._held_refs:
            del self._held_refs[node_id]

    #
    # GtkTreeModel Interface Implementation
    #
    @handle_exception(0)
    def do_get_flags(self):
        """Internal method."""
        return self.on_get_flags()

    @handle_exception(0)
    def do_get_n_columns(self):
        """Internal method."""
        return self.on_get_n_columns()

    @handle_exception(GObject.TYPE_INVALID)
    def do_get_column_type(self, index):
        """Internal method."""
        return self.on_get_column_type(index)

    @handle_exception((False, None))
    def do_get_iter(self, path):
        """Internal method."""
        return self._create_tree_iter(self.on_get_iter(path))

    @handle_exception(False)
    def do_iter_next(self, iter):
        """Internal method."""
        if iter is None:
            next_data = self.on_iter_next(None)
        else:
            next_data = self.on_iter_next(self.get_user_data(iter))

        self.set_user_data(iter, next_data)
        return next_data is not None

    @handle_exception(None)
    def do_get_path(self, iter):
        """Internal method."""
        path = self.on_get_path(self.get_user_data(iter))
        if path is None:
            return None
        else:
            return Gtk.TreePath(path)

    @handle_exception(None)
    def do_get_value(self, iter, column):
        """Internal method."""
        return self.on_get_value(self.get_user_data(iter), column)

    @handle_exception((False, None))
    def do_iter_children(self, parent):
        """Internal method."""
        data = self.get_user_data(parent) if parent else None
        return self._create_tree_iter(self.on_iter_children(data))

    @handle_exception(False)
    def do_iter_has_child(self, parent):
        """Internal method."""
        return self.on_iter_has_child(self.get_user_data(parent))

    @handle_exception(0)
    def do_iter_n_children(self, iter):
        """Internal method."""
        if iter is None:
            return self.on_iter_n_children(None)
        return self.on_iter_n_children(self.get_user_data(iter))

    @handle_exception((False, None))
    def do_iter_nth_child(self, parent, n):
        """Internal method."""
        if parent is None:
            data = self.on_iter_nth_child(None, n)
        else:
            data = self.on_iter_nth_child(self.get_user_data(parent), n)
        return self._create_tree_iter(data)

    @handle_exception((False, None))
    def do_iter_parent(self, child):
        """Internal method."""
        return self._create_tree_iter(
            self.on_iter_parent(self.get_user_data(child)))

    @handle_exception(None)
    def do_ref_node(self, iter):
        self.on_ref_node(self.get_user_data(iter))

    @handle_exception(None)
    def do_unref_node(self, iter):
        self.on_unref_node(self.get_user_data(iter))

    #
    # Python Subclass Overridables
    #
    def on_get_flags(self):
        """Overridable.

        :Returns Gtk.TreeModelFlags:
            The flags for this model. See: Gtk.TreeModelFlags
        """
        raise NotImplementedError

    def on_get_n_columns(self):
        """Overridable.

        :Returns:
            The number of columns for this model.
        """
        raise NotImplementedError

    def on_get_column_type(self, index):
        """Overridable.

        :Returns:
            The column type for the given index.
        """
        raise NotImplementedError

    def on_get_iter(self, path):
        """Overridable.

        :Returns:
            A python object (node) for the given TreePath.
        """
        raise NotImplementedError

    def on_iter_next(self, node):
        """Overridable.

        :Parameters:
            node : object
                Node at current level.

        :Returns:
            A python object (node) following
            the given node at the current level.
        """
        raise NotImplementedError

    def on_get_path(self, node):
        """Overridable.

        :Returns:
            A TreePath for the given node.
        """
        raise NotImplementedError

    def on_get_value(self, node, column):
        """Overridable.

        :Parameters:
            node : object
            column : int
                Column index to get the value from.

        :Returns:
            The value of the column for the given node."""
        raise NotImplementedError

    def on_iter_children(self, parent):
        """Overridable.

        :Returns:
            The first child of parent or None if parent has no children.
            If parent is None, return the first node of the model.
        """
        raise NotImplementedError

    def on_iter_has_child(self, node):
        """Overridable.

        :Returns:
            True if the given node has children.
        """
        raise NotImplementedError

    def on_iter_n_children(self, node):
        """Overridable.

        :Returns:
            The number of children for the given node. If node is None,
            return the number of top level nodes.
        """
        raise NotImplementedError

    def on_iter_nth_child(self, parent, n):
        """Overridable.

        :Parameters:
            parent : object
            n : int
                Index of child within parent.

        :Returns:
            The child for the given parent index starting at 0. If parent None,
            return the top level node corresponding to "n".
            If "n" is larger then available nodes, return None.
        """
        raise NotImplementedError

    def on_iter_parent(self, child):
        """Overridable.

        :Returns:
            The parent node of child or None if child is a top level node."""
        raise NotImplementedError

    def on_ref_node(self, node):
        pass

    def on_unref_node(self, node):
        pass



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import http.client
import logging
import os
import socket

try:
    from http import HTTPStatus
except ImportError:
    from http import client as HTTPStatus

from functools import partial

from gnuhealth import bus, device_cookie, fingerprints
from gnuhealth.config import CONFIG, get_config_dir
from gnuhealth.exceptions import GNUHealthServerError, \
     GNUHealthServerUnavailable
from gnuhealth.jsonrpc import Fault, ServerPool, ServerProxy

logger = logging.getLogger(__name__)
CONNECTION = None
_USER = None
CONTEXT = {}
_VIEW_CACHE = {}
_TOOLBAR_CACHE = {}
_KEYWORD_CACHE = {}
_CA_CERTS = os.path.join(get_config_dir(), 'ca_certs')
if not os.path.isfile(_CA_CERTS):
    _CA_CERTS = None

ServerProxy = partial(ServerProxy, fingerprints=fingerprints,
                      ca_certs=_CA_CERTS)
ServerPool = partial(ServerPool, fingerprints=fingerprints,
                     ca_certs=_CA_CERTS)


def context_reset():
    CONTEXT.clear()
    CONTEXT['client'] = bus.ID


context_reset()


def db_list(host, port):
    try:
        connection = ServerProxy(host, port)
        logger.info('common.db.list()')
        result = connection.common.db.list()
        logger.debug('%r', result)
        return result
    except Fault as exception:
        logger.debug(exception.faultCode)
        if exception.faultCode == str(HTTPStatus.FORBIDDEN.value):
            return []
        else:
            return None


def server_version(host, port):
    try:
        connection = ServerProxy(host, port)
        logger.info('common.server.version(None, None)')
        result = connection.common.server.version()
        logger.debug('%r', result)
        return result
    except Exception as e:
        logger.exception(e)
        return None


def authentication_services(host, port):
    try:
        connection = ServerProxy(host, port)
        logger.info('common.authentication.services()')
        services = connection.common.authentication.services()
        logger.debug('%r', services)
        return connection.url, services
    except Exception as e:
        logger.exception(e)
        return '', []


def set_service_session(parameters):
    from gnuhealth import common
    global CONNECTION, _USER
    host = CONFIG['login.host']
    hostname = common.get_hostname(host)
    port = common.get_port(host)
    database = CONFIG['login.db']
    CONFIG['login.login'] = username = parameters.get('login', [''])[0]
    try:
        user_id = int(parameters.get('user_id', [None])[0])
    except TypeError:
        pass
    session = parameters.get('session', [''])[0]
    if 'renew' in parameters:
        renew_id = int(parameters.get('renew', [-1])[0])
        if _USER != renew_id:
            raise ValueError
    _USER = user_id
    session = ':'.join(map(str, [username, user_id, session]))
    if CONNECTION is not None:
        CONNECTION.close()
    CONNECTION = ServerPool(
        hostname, port, database, session=session, cache=not CONFIG['dev'])
    bus.listen(CONNECTION)


def login(parameters):
    from gnuhealth import common
    global CONNECTION, _USER
    host = CONFIG['login.host']
    hostname = common.get_hostname(host)
    port = common.get_port(host)
    database = CONFIG['login.db']
    username = CONFIG['login.login']
    language = CONFIG['client.lang']
    parameters['device_cookie'] = device_cookie.get()
    connection = ServerProxy(hostname, port, database)
    logger.info('common.db.login(%s, %s, %s)', username, 'x' * 10, language)
    result = connection.common.db.login(username, parameters, language)
    logger.debug('%r', result)
    _USER = result[0]
    session = ':'.join(map(str, [username] + result))
    if CONNECTION is not None:
        CONNECTION.close()
    CONNECTION = ServerPool(
        hostname, port, database, session=session, cache=not CONFIG['dev'])
    device_cookie.renew()
    bus.listen(CONNECTION)


def logout():
    global CONNECTION, _USER
    if CONNECTION is not None:
        try:
            logger.info('common.db.logout()')
            with CONNECTION() as conn:
                conn.common.db.logout()
        except (Fault, socket.error, http.client.CannotSendRequest):
            pass
        CONNECTION.close()
        CONNECTION = None
    _USER = None


def execute(*args):
    global CONNECTION, _USER
    if CONNECTION is None:
        raise GNUHealthServerError('403')
    try:
        name = '.'.join(args[:3])
        args = args[3:]
        logger.info('%s%r', name, args)
        with CONNECTION() as conn:
            result = getattr(conn, name)(*args)
    except (http.client.CannotSendRequest, socket.error) as exception:
        raise GNUHealthServerUnavailable(*exception.args)
    logger.debug('%r', result)
    return result


def clear_cache(prefix=None):
    if CONNECTION:
        CONNECTION.clear_cache(prefix)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from gi.repository import Gtk

from gnuhealth.common.entry_position import reset_position
from gnuhealth.common.number_entry import NumberEntry

from .widget import Widget


class Integer(Widget):
    "Integer"

    def __init__(self, view, attrs):
        super(Integer, self).__init__(view, attrs)
        self.widget = Gtk.HBox()
        self.entry = self.mnemonic_widget = NumberEntry(
            monetary=attrs.get('monetary', False))
        self.entry.props.activates_default = True
        self.entry.connect('activate', self.sig_activate)
        self.entry.connect('focus-out-event', lambda *a: self._focus_out())
        self.entry.connect('key-press-event', self.send_modified)
        self.symbol = attrs.get('symbol')
        if self.symbol:
            self.symbol_start = Gtk.Entry(editable=False)
            self.widget.pack_start(
                self.symbol_start, expand=False, fill=False, padding=1)
        self.widget.pack_start(self.entry, expand=False, fill=False, padding=0)
        self.factor = float(attrs.get('factor', 1))
        self.grouping = bool(int(attrs.get('grouping', 1)))
        if self.symbol:
            self.symbol_end = Gtk.Entry(editable=False)
            self.widget.pack_start(
                self.symbol_end, expand=False, fill=False, padding=1)

    @property
    def modified(self):
        if self.record and self.field:
            value = self.get_client_value()
            return value != self.get_value()
        return False

    def set_value(self):
        return self.field.set_client(
            self.record, self.entry.get_text(), factor=self.factor)

    def get_value(self):
        return self.entry.get_text()

    def get_client_value(self):
        if not self.field:
            value = ''
        else:
            value = self.field.get_client(
                self.record, factor=self.factor, grouping=self.grouping)
        return value

    @property
    def width(self):
        return self.attrs.get('width', 8)

    def display(self):
        def set_symbol(entry, text):
            if text:
                entry.set_text(text)
                entry.set_width_chars(len(text) + 1)
                entry.show()
            else:
                entry.set_text('')
                entry.hide()
        super().display()
        self.entry.set_width_chars(self.width)
        if self.field and self.symbol:
            symbol, position = self.field.get_symbol(self.record, self.symbol)
            if position < 0.5:
                set_symbol(self.symbol_start, symbol)
                set_symbol(self.symbol_end, '')
            else:
                set_symbol(self.symbol_start, '')
                set_symbol(self.symbol_end, symbol)
        value = self.get_client_value()
        self.entry.set_text(value)
        reset_position(self.entry)

    def _readonly_set(self, value):
        super()._readonly_set(value)
        self.entry.set_editable(not value)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later


from gnuhealth.config import GNUHEALTH_ICON
from gi.repository import Gtk


class Activity():
    "GNU Health client Activity Logger"
    activity_window = Gtk.Window()
    activity_window.set_default_size(500, 500)
    activity_window.set_title("Activity log - GNU Health ")
    activity_window.set_icon(GNUHEALTH_ICON)

    sw = Gtk.ScrolledWindow()
    sw.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

    # TextView
    activity = Gtk.TextView()
    sw.add(activity)

    # Make it read-only
    activity.set_editable(False)
    textbuffer = activity.get_buffer()

    activity_window.add(sw)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import datetime as dt
from unittest import TestCase

from gnuhealth.common.timedelta import DEFAULT_CONVERTER, format, parse


class TimeDeltaTestCase(TestCase):
    "Test common timedelta"

    def _format_values(self):
        return [
            (None, ''),
            (dt.timedelta(), '00:00'),
            (dt.timedelta(days=3, hours=5, minutes=30), '3d 05:30'),
            (dt.timedelta(weeks=48), '11M 6d'),
            (dt.timedelta(weeks=50), '11M 2w 6d'),
            (dt.timedelta(weeks=52), '12M 4d'),
            (dt.timedelta(days=360), '12M'),
            (dt.timedelta(days=364), '12M 4d'),
            (dt.timedelta(days=365), '1Y'),
            (dt.timedelta(days=366), '1Y 1d'),
            (dt.timedelta(hours=2, minutes=5, seconds=10), '02:05:10'),
            (dt.timedelta(minutes=15, microseconds=42), '00:15:00.000042'),
            (dt.timedelta(days=1, microseconds=42), '1d .000042'),
            (dt.timedelta(seconds=-1), '-00:00:01'),
            (dt.timedelta(days=-1, hours=-5, minutes=-30), '-1d 05:30'),
        ]

    def _time_only_converter(self):
        converter = {}
        converter['s'] = DEFAULT_CONVERTER['s']
        converter['m'] = DEFAULT_CONVERTER['m']
        converter['h'] = DEFAULT_CONVERTER['h']
        converter['d'] = 0
        return converter

    def _time_only_converter_values(self):
        return [
            (None, ''),
            (dt.timedelta(), '00:00'),
            (dt.timedelta(days=5, hours=5, minutes=30), '125:30'),
            (dt.timedelta(hours=2, minutes=5, seconds=10), '02:05:10'),
            (dt.timedelta(minutes=15, microseconds=42), '00:15:00.000042'),
            (dt.timedelta(days=1, microseconds=42), '24:00:00.000042'),
            (dt.timedelta(seconds=-1), '-00:00:01'),
            (dt.timedelta(days=-1, hours=-5, minutes=-30), '-29:30'),
        ]

    def test_format(self):
        "Test format"
        for timedelta, text in self._format_values():
            self.assertEqual(format(timedelta), text,
                             msg="format(%r)" % timedelta)

    def test_format_time_only_converter(self):
        "Test format with time only converter"
        converter = self._time_only_converter()
        for timedelta, text in self._time_only_converter_values():
            self.assertEqual(format(timedelta, converter), text,
                             msg="format(%r)" % timedelta)

    def _parse_values(self):
        return self._format_values() + [
            (dt.timedelta(), '  '),
            (dt.timedelta(), 'foo'),
            (dt.timedelta(days=1.5), '1.5d'),
            (dt.timedelta(days=-2), '1d -1d'),
            (dt.timedelta(hours=1, minutes=5, seconds=10), '1:5:10:42'),
            (dt.timedelta(hours=2), '1: 1:'),
            (dt.timedelta(hours=.25), ':15'),
            (dt.timedelta(hours=1), '1h'),
            (dt.timedelta(hours=.25), '.25h'),
        ]

    def test_parse(self):
        "Test parse"
        for timedelta, text in self._parse_values():
            self.assertEqual(parse(text), timedelta,
                             msg="parse(%r)" % text)

    def test_parse_time_only_converter(self):
        "Test parse with time only converter"
        converter = self._time_only_converter()
        for timedelta, text in self._time_only_converter_values():
            self.assertEqual(parse(text, converter), timedelta,
                             msg="parse(%r)" % text)



#########################################################################
#             GNU HEALTH HOSPITAL MANAGEMENT - GTK CLIENT               #
#                      https://www.gnuhealth.org                        #
#########################################################################
#       The GNUHealth HMIS client based on the Tryton GTK Client        #
#########################################################################
#
# SPDX-FileCopyrightText: 2008-2024 The Tryton Community <info@tryton.org>
# SPDX-FileCopyrightText: 2017-2024 GNU Health Community <info@gnuhealth.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

# This file is part of GNU Health.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
import gettext
import webbrowser
from urllib.parse import urlencode, urljoin

from gi.repository import Gtk

import gnuhealth.common as common
from gnuhealth.common.common import selection
from gnuhealth.common.underline import set_underline
from gnuhealth.config import CONFIG
from gnuhealth.rpc import CONNECTION

from .char import Char
from .widget import TranslateMixin, Widget

_ = gettext.gettext


class URL(Char):
    "url"

    def __init__(self, view, attrs):
        super(URL, self).__init__(view, attrs)

        self.tooltips = common.Tooltips()
        self.button = Gtk.Button()
        self.button.set_image(common.IconFactory.get_image(
            'gnuhealth-public', Gtk.IconSize.SMALL_TOOLBAR))
        self.button.set_relief(Gtk.ReliefStyle.NONE)
        self.button.connect('clicked', self.button_clicked)
        self.widget.pack_start(
            self.button, expand=False, fill=False, padding=0)

    def display(self):
        super(URL, self).display()
        self.set_tooltips()
        if self.record and 'icon' in self.attrs:
            icon = self.attrs['icon']
            if icon in self.record.group.fields:
                value = self.record[icon].get_client(self.record)
                value = value if value else 'gnuhealth-public'
            else:
                value = icon
            self.button.set_image(common.IconFactory.get_image(
                value, Gtk.IconSize.SMALL_TOOLBAR))
        self.button.set_sensitive(bool(self.entry.get_text()))

    def set_tooltips(self):
        value = self.entry.get_text()
        if value:
            self.tooltips.enable()
            self.tooltips.set_tip(self.button, value)
        else:
            self.tooltips.set_tip(self.button, '')
            self.tooltips.disable()

    def button_clicked(self, widget):
        value = self.entry.get_text()
        if value:
            webbrowser.open(value, new=2)


class Email(URL):
    "email"

    def button_clicked(self, widget):
        value = self.entry.get_text()
        if value:
            webbrowser.open('mailto:%s' % value, new=2)

    def set_tooltips(self):
        value = self.entry.get_text()
        if value:
            self.tooltips.enable()
            self.tooltips.set_tip(self.button, 'mailto:%s' % value)
        else:
            self.tooltips.set_tip(self.button, '')
            self.tooltips.disable()


class CallTo(URL):
    "call to"

    def button_clicked(self, widget):
        value = self.entry.get_text()
        if value:
            webbrowser.open('callto:%s' % value, new=2)

    def set_tooltips(self):
        value = self.entry.get_text()
        if value:
            self.tooltips.enable()
            self.tooltips.set_tip(self.button, 'callto:%s' % value)
        else:
            self.tooltips.set_tip(self.button, '')
            self.tooltips.disable()


class SIP(URL):
    "sip"

    def button_clicked(self, widget):
        value = self.entry.get_text()
        if value:
            webbrowser.open('sip:%s' % value, new=2)

    def set_tooltips(self):
        value = self.entry.get_text()
        if value:
            self.tooltips.enable()
            self.tooltips.set_tip(self.button, 'sip:%s' % value)
        else:
            self.tooltips.set_tip(self.button, '')
            self.tooltips.disable()


class HTML(Widget, TranslateMixin):
    "HTML"

    def __init__(self, view, attrs):
        super().__init__(view, attrs)
        self.widget = Gtk.HBox()
        self.mnemonic_widget = self.button = Gtk.LinkButton()
        self.button.set_label(set_underline(attrs['string']))
        self.button.set_use_underline(True)
        self.button.set_alignment(0, 0.5)
        self.widget.pack_start(
            self.button, expand=False, fill=False, padding=0)

        if attrs.get('translate'):
            self.widget.pack_start(
                self.translate_button(), expand=False, fill=False, padding=0)

    def uri(self, language=None):
        if not self.record or self.record.id < 0 or CONNECTION.url is None:
            uri = ''
        else:
            path = ['ir/html', self.model_name, str(self.record.id),
                    self.field_name]
            params = {
                'language': language or CONFIG['client.lang'],
                'title': CONFIG['client.title'],
            }
            uri = urljoin(
                CONNECTION.url + '/', '/'.join(path) + '?' + urlencode(params))
        return uri

    def display(self):
        super().display()
        self.button.set_uri(self.uri())

    def _readonly_set(self, value):
        super()._readonly_set(value)
        for button in self.widget.get_children():
            button.set_sensitive(not value)

    def translate_dialog(self, languages):
        languages = {lang['name']: lang['code'] for lang in languages}
        result = selection(_('Choose a language'), languages)
        if result:
            webbrowser.open(self.uri(language=result[1]), new=2)
